#!/usr/bin/env python
'''Export python grammar to C grammar files

Author: Jeroen van der Heijden (Transceptor Technology)
Date: 2016-10-10
'''
import sys
import os
from grammar import siri_grammar
from pyleri import Grammar


if __name__ == '__main__':
    c_file, h_file = siri_grammar.export_c(target='siri/grammar/grammar')

    EXPORT_PATH = 'cgrammar'

    try:
        os.makedirs(EXPORT_PATH)
    except FileExistsError:
        pass

    with open(os.path.join(EXPORT_PATH, 'grammar.c'),
              'w',
              encoding='utf-8') as f:
        f.write(c_file)
    with open(os.path.join(EXPORT_PATH, 'grammar.h'),
              'w',
              encoding='utf-8') as f:
        f.write(h_file)

    print('\nFinished creating new c-grammar files...\n')

    EXPORT_PATH = 'jsgrammar'

    try:
        os.makedirs(EXPORT_PATH)
    except FileExistsError:
        pass

    js_file = siri_grammar.export_js(js_template=Grammar.JS_WINDOW_TEMPLATE)

    with open(os.path.join(EXPORT_PATH, 'grammar.js'),
              'w',
              encoding='utf-8') as f:
        f.write(js_file)

    js_es6_file = siri_grammar.export_js()

    with open(os.path.join(EXPORT_PATH, 'SiriGrammar.js'),
              'w',
              encoding='utf-8') as f:
        f.write(js_es6_file)

    print('\nFinished creating new js-grammar files...\n')

    py_file = siri_grammar.export_py()

    EXPORT_PATH = 'pygrammar'

    try:
        os.makedirs(EXPORT_PATH)
    except FileExistsError:
        pass

    with open(os.path.join(EXPORT_PATH, 'grammar.py'),
              'w',
              encoding='utf-8') as f:
        f.write(py_file)

    print('\nFinished creating new py-grammar file...\n')

    go_file = siri_grammar.export_go()

    EXPORT_PATH = 'gogrammar'

    try:
        os.makedirs(EXPORT_PATH)
    except FileExistsError:
        pass

    with open(os.path.join(EXPORT_PATH, 'grammar.go'),
              'w',
              encoding='utf-8') as f:
        f.write(go_file)

    print('\nFinished creating new go-grammar file...\n')

    java_file = siri_grammar.export_java()

    EXPORT_PATH = 'javagrammar'

    try:
        os.makedirs(EXPORT_PATH)
    except FileExistsError:
        pass

    with open(os.path.join(EXPORT_PATH, 'SiriGrammar.java'),
              'w',
              encoding='utf-8') as f:
        f.write(java_file)

    print('\nFinished creating new java-grammar file...\n')



import os
import logging
import random
import asyncio
from .constants import SERVICE

VERBOSE = ' --verbose'


class SiriDB:

    LOG_LEVEL = 'CRITICAL'

    def __init__(self,
                 dbname='dbtest',
                 time_precision='s',
                 buffer_path='',
                 duration_log='1d',
                 duration_num='1w',
                 buffer_size=1024,
                 **unused):
        self.dbname = dbname
        self.time_precision = time_precision
        self.buffer_path = buffer_path
        self.duration_log = duration_log
        self.duration_num = duration_num
        self.buffer_size = buffer_size
        self.servers = []

    async def create_on(self, server, sleep=None):
        await asyncio.sleep(1)

        logging.info('Create database {} on {}'.format(
            self.dbname,
            server.name))

        rc = os.system(
            '{service} '
            '-u sa -p siri -s {addr} '
            'new-database '
            '--db-name {dbname} '
            '--time-precision {time_precision} '
            '--duration-log {duration_log} '
            '--duration-num {duration_num} '
            '--buffer-size {buffer_size}'
            '{verbose}'.format(
                service=SERVICE,
                addr=server.addr,
                verbose=VERBOSE if self.LOG_LEVEL == 'DEBUG'
                else ' >/dev/null',
                **vars(self)))

        assert rc == 0, 'Expected rc = 0 but got rc = {}'.format(rc)

        self.servers.append(server)

        if sleep:
            await asyncio.sleep(sleep)

    async def drop_db(
                self,
                server,
                username='iris',
                password='siri',
                sleep=None):

        rc = os.system(
            '{service} '
            '-u sa -p siri -s {addr} '
            'drop-database '
            '--db-name {dbname} '
            '--ignore-offline '
            '--force{verbose}'.format(
                service=SERVICE,
                addr=server.addr,
                verbose=VERBOSE if self.LOG_LEVEL == 'DEBUG'
                else ' >/dev/null',
                **vars(self)))

        assert rc == 0, 'Expected rc = 0 but got rc = {}'.format(rc)

        self.servers.append(server)

        if sleep:
            await asyncio.sleep(sleep)

    async def add_replica(
                self,
                server,
                pool,
                username='iris',
                password='siri',
                remote_server=None,
                sleep=None):

        if remote_server is None:
            remote_server = random.choice(self.servers)

        rc = os.system(
            '{service} '
            '-u sa -p siri -s {addr} '
            'new-replica '
            '--db-name {dbname} '
            '--db-server {dbaddr} '
            '--db-user {dbuser} '
            '--db-password {dbpassword} '
            '--pool {pool} '
            '--force{verbose}'.format(
                service=SERVICE,
                addr=server.addr,
                dbaddr=remote_server.addr,
                dbuser=username,
                dbpassword=password,
                pool=pool,
                verbose=VERBOSE if self.LOG_LEVEL == 'DEBUG'
                else ' >/dev/null',
                **vars(self)))

        assert rc == 0, 'Expected rc = 0 but got rc = {}'.format(rc)

        self.servers.append(server)

        if sleep:
            await asyncio.sleep(sleep)

    async def add_pool(
                self,
                server,
                username='iris',
                password='siri',
                remote_server=None,
                sleep=None):

        if remote_server is None:
            remote_server = random.choice(self.servers)

        rc = os.system(
            '{service} '
            '-u sa -p siri -s {addr} '
            'new-pool '
            '--db-name {dbname} '
            '--db-server {dbaddr} '
            '--db-user {dbuser} '
            '--db-password {dbpassword} '
            '--force{verbose}'.format(
                service=SERVICE,
                addr=server.addr,
                dbaddr=remote_server.addr,
                dbuser=username,
                dbpassword=password,
                verbose=VERBOSE if self.LOG_LEVEL == 'DEBUG'
                else ' >/dev/null',
                **vars(self)))

        assert rc == 0, 'Expected rc = 0 but got rc = {}'.format(rc)

        self.servers.append(server)

        if sleep:
            await asyncio.sleep(sleep)



import re
import random
import asyncio
import functools
import random
import time
import os
import sys
from math import nan
from testing import default_test_setup
from testing import parse_args
from testing import run_test
from testing import Server
from testing import SiriDB
from testing import TestBase
from testing.constants import PYGRAMMAR_PATH
from querygenerator.querygenerator import QueryGenerator
from querygenerator.k_map import k_map
sys.path.append(PYGRAMMAR_PATH)
from grammar import SiriGrammar  # nopep8


def gen_simple_data(m, n):
    series = {
        str(a).zfill(6): [[b*n+a, random.randint(0, 20)] for b in range(n)]
        for a in range(m)}
    return series


def update_k_map_show(show):
    kv = {a['name']: a['value'] for a in show['data']}
    k_map['r_integer']['k_active_handles'] = kv['active_handles']
    k_map['r_doubleq_str']['k_buffer_path'] = '"'+kv['buffer_path']+'"'
    k_map['r_integer']['k_buffer_size'] = kv['buffer_size']
    k_map['r_doubleq_str']['k_dbpath'] = '"'+kv['dbpath']+'"'
    k_map['r_float']['k_drop_threshold'] = kv['drop_threshold']
    k_map['r_doubleq_str']['k_ip_support'] = '"'+kv['ip_support']+'"'
    k_map['r_doubleq_str']['k_libuv'] = '"'+kv['libuv']+'"'
    k_map['r_integer']['k_max_open_files'] = kv['max_open_files']
    k_map['r_integer']['k_mem_usage'] = kv['mem_usage']
    k_map['r_integer']['k_open_files'] = kv['open_files']
    k_map['r_integer']['k_pool'] = kv['pool']
    k_map['r_integer']['k_received_points'] = kv['received_points']
    k_map['r_uinteger']['k_list_limit'] = kv['list_limit']
    k_map['r_integer']['k_startup_time'] = kv['startup_time']
    k_map['r_doubleq_str']['k_status'] = '"'+kv['status']+'"'
    k_map['r_doubleq_str']['k_sync_progress'] = '"'+kv['sync_progress']+'"'
    k_map['r_doubleq_str']['k_timezone'] = '"'+kv['timezone']+'"'
    k_map['r_integer']['k_uptime'] = kv['uptime']
    k_map['r_uuid_str']['r_uuid_str'] = kv['uuid']
    k_map['r_doubleq_str']['k_server'] = '"'+kv['server']+'"'
    k_map['r_doubleq_str']['uuid'] = '"'+kv['server']+'"'
    k_map['r_doubleq_str']['k_version'] = '"'+kv['version']+'"'
    k_map['r_uinteger']['k_port'] = kv['server'].split(':', 1)[1]
    k_map['r_uinteger']['k_select_points_limit'] = \
        kv['select_points_limit']
    k_map['r_doubleq_str']['k_reindex_progress'] = \
        '"'+kv['reindex_progress']+'"'


class TestGrammar(TestBase):
    title = 'Test from grammar'

    async def test_create_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': ''
            }})
        for q in qb.generate_queries('create_stmt'):
            await self.client0.query(q)

    async def test_select_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': '',
                'k_filter': '',  # skip because only number type series
                'k_prefix': '',  # skip
                'k_suffix': '',  # skip
                'k_merge': '',  # skip
                'k_where': '',  # skip
                'after_expr': '',  # skip
                'before_expr': '',  # skip
                'between_expr': '',  # skip
                }
            })
        for q in qb.generate_queries('select_stmt'):
            await self.client0.query(q)

    async def test_revoke_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': ''
            }})
        for q in qb.generate_queries('revoke_stmt'):
            await self.client0.query(q)

    async def test_grant_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': ''
            }})
        for q in qb.generate_queries('grant_stmt'):
            await self.client0.query(q)

    async def test_alter_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': '', 
                'k_now': '',  # not possible (set expiration num/log)
                'set_name': '',  # skip
                'set_address': '',  # not possible
                'set_port': '',  # not possible
                'set_timezone': '',  # same value error
                'set_log_level': '',  # not required, but skip to keep log level
            }})
        for q in qb.generate_queries('alter_stmt'):
            await self.client0.query(q)

    async def test_count_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': ''
            }})
        for q in qb.generate_queries('count_stmt'):
            await self.client0.query(q)

    async def test_list_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': ''
            }})
        for q in qb.generate_queries('list_stmt'):
            await self.client0.query(q)

    async def test_drop_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': '',
                'drop_server': '',   # not possible
            }})
        for q in qb.generate_queries('drop_stmt'):
            await self.client0.query(q)

    async def test_show_stmt(self):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': ''
            }})
        for q in qb.generate_queries('show_stmt'):
            await self.client0.query(q)

    @default_test_setup(1)
    async def run(self):
        await self.client0.connect()

        # await self.db.add_pool(self.server1, sleep=2)

        update_k_map_show(await self.client0.query('show'))

        series = gen_simple_data(20, 70)

        await self.client0.insert(series)
        await self.client0.query('create group `GROUP_OR_TAG` for /00000.*/')
        
        await self.test_create_stmt()

        time.sleep(2)

        await self.test_count_stmt()

        await self.test_list_stmt()

        await self.test_select_stmt()

        await self.test_revoke_stmt()

        await self.test_grant_stmt()

        await self.test_alter_stmt()

        await self.test_drop_stmt()

        await self.test_show_stmt()

        self.client0.close()

        return False

class TestGrammarStart(TestBase):

    async def test_all_stmts(self, client):
        qb = QueryGenerator(SiriGrammar, {
            'regex_map': k_map,
            'replace_map': {
                'r_singleq_str': '',
                'r_comment': '',
                'k_timeit': '',
                'select_stmt': '',
                'list_stmt': '',
                'count_stmt': '',

                'alter_group': '',
                #'drop_group': '',
                'alter_server': '',
                'drop_server': '',
                'alter_user': '',
                'drop_user': '',

                #'set_address': '',
                #'set_port': '',
                'set_timezone': '',
                'set_log_level': '',  # not required, skip to keep log level
                'set_expiration_num': '',
                'set_expiration_log': '',

                'k_prefix': '',
                'k_suffix': '',
                'k_filter': '',
                #'k_where': '',
                'after_expr': '',
                'before_expr': '',
                'between_expr': '',
                'k_merge': '',
        }})
        for q in qb.generate_queries():
            await self.client0.query(q)

    @default_test_setup(1)
    async def run(self):
        await self.client0.connect()

        # await self.db.add_pool(self.server1, sleep=2)

        update_k_map_show(await self.client0.query('show'))

        series = gen_simple_data(20, 70)

        await self.client0.insert(series)
        await self.client0.query('create group `GROUP_OR_TAG` for /00000.*/')
        #time.sleep(2)
        await self.test_all_stmts()
        self.client0.close()
        return False


if __name__ == '__main__':
    parse_args()
    run_test(TestGrammar())



import os
import sys


class QueryGenerator(list):

    def __init__(self, grammar, maps={}):
        self.grammar = grammar
        self.ignores = maps.get('replace_map', {})
        self.max_repeat_n_map = maps.get('max_repeat_n_map', {})
        self.max_list_n_map = maps.get('max_list_n_map', {})
        self.default_list_n = maps.get('default_list_n', 1)
        self.regex_map = maps.get('regex_map', {})

    def generate_queries(self, ename='START'):
        try:
            ele = getattr(self.grammar, ename)
        except AttributeError:
            print('{} element not found in tree'.format(ename))
        else:
            for q in self._addq([ele]):
                yield ' '.join(map(str, q)).strip()

    def _addq(self, q):
        for i, e in enumerate(q):
            if isinstance(e, str) or isinstance(e, int) or \
                    isinstance(e, float):
                continue

            ename = getattr(e, 'name', None)
            iv = self.ignores.get(ename)
            if iv is not None:
                q[i] = iv
                break

            self.append(ename)
            func = getattr(self, '_on_'+e.__class__.__name__, lambda q, i: [])
            for q1 in func(q, i):
                for q2 in self._addq(q1):
                    yield q2
            self.pop()
            q[i] = e
            break
        else:
            yield q

    def _on_Keyword(self, q, i):
        q[i] = q[i]._keyword
        yield q

    def _on_Regex(self, q, i):
        regex_ename = self[-1]
        if regex_ename in self.regex_map:
            re_map = self.regex_map[regex_ename]
            for ename in self[::-1]:
                val = re_map.get(ename)
                if val is not None:
                    q[i] = val
                    yield q
                    break
        
        # debug code
        
        #     else:
        #         print('no value found for:')
        #         print(ename)
        #         print(q[:i+1])
        #         print(i, len(self), self)
        #         print(self.regex_map)
        # else:
        #     print('unknown regex element')
        #     print(ename)


    def _on_Token(self, q, i):
        q[i] = q[i]._token
        yield q

    def _on_Tokens(self, q, i):
        for e1 in q[i]._tokens:
            q[i] = e1
            yield q

    def _on_Sequence(self, q, i):
        q = q[:i]+q[i]._elements+q[i+1:]
        yield q

    def _on_List(self, q, i):
        if q[i]._min == 0:
            q0 = q[:i]+q[i+1:]
            yield q0

        if q[i]._max == 1:
            q0 = q[:i]+list(q[i]._elements)[:1]+q[i+1:]
            yield q0
        else:
            name = getattr(q[i], 'name', getattr(q[i]._element, 'name', None))
            n = self.max_list_n_map.get(name, self.default_list_n)
            if q[i]._max:
                n = min(n, q[i]._max)

            eles = list(q[i]._elements)[:1] + \
                [q[i]._delimiter, list(q[i]._elements)[0]]*(n-1)

            q0 = q[:i]+eles+q[i+1:]
            yield q0

    def _on_Repeat(self, q, i):
        if q[i]._min == 0:
            q0 = q[:i]+q[i+1:]
            yield q0

        if q[i]._max == 1:
            q0 = q[:i]+list(q[i]._elements)[:1]+q[i+1:]
            yield q0
        else:
            n = self.max_repeat_n_map.get(getattr(q[i], 'name', None), 1)
            eles = list(q[i]._elements)[:1]*n
            q0 = q[:i]+eles+q[i+1:]
            yield q0

    def _on_Optional(self, q, i):
        q[i] = list(q[i]._elements)[0]
        yield q
        q[i] = ''
        yield q

    def _on_Choice(self, q, i):
        for e1 in q[i]._elements:
            q[i] = e1
            yield q

    def _on_Rule(self, q, i):
        for e1 in q[i]._element._elements:
            if e1.__class__.__name__ == 'Sequence' and sum(
                    e2.__class__.__name__ == 'This' for e2 in e1._elements):
                continue

            q[i] = e1
            yield q



'''Help module.

Build help structure for SiriDB grammar.

:copyright: 2016, Jeroen van der Heijden (Transceptor Technology)
'''

import os
import re
import sys

#
# Created with ls -A1
#
help_files = [f[:-3] for f in os.listdir('../help') if f.endswith('.md')]
help_files.sort()

def _build_structure(help_files):
    '''Return a tree structure for the help files.'''
    _structure = {}

    def _walk(d, keys, i):
        i += 1
        k = tuple(keys[:i])
        if k not in d:
            d[k] = {}
        if i < len(keys):
            _walk(d[k], keys, i)
    for help_file in help_files:
        _walk(_structure, help_file.split('_'), 0)
    return _structure


help_structure = _build_structure(help_files)


if __name__ == '__main__':
    print(help_structure)


TEST_DIR = './testdir'
PYGRAMMAR_PATH = '../grammar'
SIRIDBC = '../{BUILDTYPE}/siridb-server'
SERVICE = '/usr/local/bin/siridb-admin'
VALGRIND = 'valgrind' \
  ' --tool=memcheck' \
  ' --error-exitcode=1' \
  ' --leak-check=full' \
  ' --show-leak-kinds=all' \
  ' --track-origins=yes' \
  ' -v '
VALGRIND = 'valgrind' \
  ' --tool=memcheck'  \
  ' --error-exitcode=1' \
  ' --leak-check=full '

MAX_OPEN_FILES = 512    # Default value is 32768 but with valgrind 512 is max



import asyncio
import functools
import random
import time
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import parse_args


TIME_PRECISION = 's'


class TestList(TestBase):
    title = 'Test list'

    GEN_POINTS = functools.partial(
        gen_points, n=1, time_precision=TIME_PRECISION)

    @default_test_setup(1, time_precision=TIME_PRECISION)
    async def run(self):
        await self.client0.connect()

        # Create some random series and start 25 insert task parallel
        series = gen_series(n=10000)
        tasks = [
            asyncio.ensure_future(
                self.client0.insert_some_series(
                    series,
                    timeout=0,
                    points=self.GEN_POINTS))
            for i in range(25)]

        await asyncio.gather(*tasks)

        await self.client0.query('list series /.*/ - /.*/')
        await self.client0.query('list series /.*/ | /.*/')
        await self.client0.query('list series /.*/ & /.*/')
        await self.client0.query('list series /.*/ ^ /.*/')

        await self.client0.query('list series /.*/ - /a.*/')
        await self.client0.query('list series /.*/ | /a.*/')
        await self.client0.query('list series /.*/ & /a.*/')
        await self.client0.query('list series /.*/ ^ /a.*/')

        await self.client0.query('list series /.*/ - /a.*/ | /b.*/')
        await self.client0.query('list series /.*/ | /a.*/ | /b.*/')
        await self.client0.query('list series /.*/ & /a.*/ | /b.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ | /b.*/')

        await self.client0.query('list series /.*/ - /a.*/ | /.*/')
        await self.client0.query('list series /.*/ | /a.*/ | /.*/')
        await self.client0.query('list series /.*/ & /a.*/ | /.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ | /.*/')

        await self.client0.query('list series /.*/ - /a.*/ - /b.*/')
        await self.client0.query('list series /.*/ | /a.*/ - /b.*/')
        await self.client0.query('list series /.*/ & /a.*/ - /b.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ - /b.*/')

        await self.client0.query('list series /.*/ - /a.*/ - /.*/')
        await self.client0.query('list series /.*/ | /a.*/ - /.*/')
        await self.client0.query('list series /.*/ & /a.*/ - /.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ - /.*/')

        await self.client0.query('list series /.*/ - /a.*/ & /b.*/')
        await self.client0.query('list series /.*/ | /a.*/ & /b.*/')
        await self.client0.query('list series /.*/ & /a.*/ & /b.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ & /b.*/')

        await self.client0.query('list series /.*/ - /a.*/ & /.*/')
        await self.client0.query('list series /.*/ | /a.*/ & /.*/')
        await self.client0.query('list series /.*/ & /a.*/ & /.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ & /.*/')

        await self.client0.query('list series /.*/ - /a.*/ ^ /b.*/')
        await self.client0.query('list series /.*/ | /a.*/ ^ /b.*/')
        await self.client0.query('list series /.*/ & /a.*/ ^ /b.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ ^ /b.*/')

        await self.client0.query('list series /.*/ - /a.*/ ^ /.*/')
        await self.client0.query('list series /.*/ | /a.*/ ^ /.*/')
        await self.client0.query('list series /.*/ & /a.*/ ^ /.*/')
        await self.client0.query('list series /.*/ ^ /a.*/ ^ /.*/')

        await self.client0.query('alter database set list_limit 5000')
        with self.assertRaisesRegex(
                QueryError,
                'Limit must be a value between 1 and 5000 '
                'but received: 6000.*'):
            await self.client0.query(
                'list series limit 6000')

        self.client0.close()


if __name__ == '__main__':
    parse_args()
    run_test(TestList())



import os
import subprocess
import argparse
from .server import Server
from .color import Color
from .constants import SIRIDBC


def is_valgrind_installed():
    with open(os.devnull, 'w') as fnull:
        try:
            subprocess.call(['valgrind'], stdout=fnull, stderr=fnull)
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                return False
    return True


def print_siridb_version(args):
    fn = SIRIDBC.format(BUILDTYPE=args.build)
    try:
        p = subprocess.Popen(
            [fn, '--version'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
    except FileNotFoundError:
        print(Color.error(f'Cannot find: {fn}'))
        exit(1)

    output, err = p.communicate()
    rc = p.returncode
    if rc or err:
        print(Color.error(f'Cannot use: {fn}'))
        exit(1)

    print(f'Test with: {Color.info(output.decode().splitlines()[0])}')


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-t', '--terminal',
        choices=['xterm', 'xfce4-terminal'],
        default=None,
        help='Start SiriDB servers in a terminal. If no terminal is given '
        'process put their output in log files.')

    parser.add_argument(
        '-m', '--mem-check',
        action='store_true',
        help='Use `valgrind` for memory errors and leaks.')

    parser.add_argument(
        '-k', '--keep',
        action='store_true',
        help='Only valid when a terminal is used. This will keep the terminal '
        'open.')

    parser.add_argument(
        '-b', '--build',
        choices=['Release', 'Debug'],
        default='Release',
        help='Choose either the Release or Debug build.')

    parser.add_argument(
        '-l', '--log-level',
        default='critical',
        help='set the log level',
        choices=['debug', 'info', 'warning', 'error', 'critical'])

    args = parser.parse_args()

    has_valgrind = is_valgrind_installed()

    print_siridb_version(args)
    print("Test using valgrind for memory errors and leaks: ", end='')
    if args.mem_check and not has_valgrind:
        print(Color.warning('disabled (!! valgrind not found !!)'))
    elif not args.mem_check:
        print(Color.warning('disabled'))
    else:
        print(Color.success('enabled'))

    Server.MEM_CHECK = args.mem_check and has_valgrind
    Server.HOLD_TERM = args.keep
    Server.TERMINAL = args.terminal
    Server.BUILDTYPE = args.build
    Server.LOG_LEVEL = args.log_level.upper()



import logging
import re
from siridbhelp import help_structure
from pyleri import Choice
from pyleri import Grammar
from pyleri import Keyword
from pyleri import List
from pyleri import Optional
from pyleri import Prio
from pyleri import Ref
from pyleri import Regex
from pyleri import Repeat
from pyleri import Sequence
from pyleri import THIS
from pyleri import Token
from pyleri import Tokens


class SiriGrammar(Grammar):
    '''
    SiriDB grammar.

    Note: choices can be optimized using most_greedy=False when there
          is a preferable order in choices.
          This only should be used when there's no conflict in making a
          decision by the parser. (e.g. two choices should start with the
          same keyword because in that case we should usually take the most
          greedy one)
    '''

    RE_KEYWORDS = re.compile('[a-z_]+')

    # Regular expressions
    r_float = Regex('[-+]?[0-9]*\.?[0-9]+')
    r_integer = Regex('[-+]?[0-9]+')
    r_uinteger = Regex('[0-9]+')
    r_time_str = Regex('[0-9]+[smhdw]')
    r_singleq_str = Regex('(?:\'(?:[^\']*)\')+')
    r_doubleq_str = Regex('(?:"(?:[^"]*)")+')
    r_grave_str = Regex('(?:`(?:[^`]*)`)+')
    r_uuid_str = Regex(
        '[0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{12}')
    # we only allow an optional 'i' for case-insensitive regex
    r_regex = Regex('(/[^/\\\\]*(?:\\\\.[^/\\\\]*)*/i?)')
    r_comment = Regex('#.*')

    # Keywords
    k_access = Keyword('access')
    k_active_handles = Keyword('active_handles')
    k_active_tasks = Keyword('active_tasks')
    k_address = Keyword('address')
    k_after = Keyword('after')
    k_all = Keyword('all')
    k_alter = Keyword('alter')
    k_and = Keyword('and')
    k_as = Keyword('as')
    k_backup_mode = Keyword('backup_mode')
    k_before = Keyword('before')
    k_buffer_size = Keyword('buffer_size')
    k_buffer_path = Keyword('buffer_path')
    k_between = Keyword('between')
    k_count = Keyword('count')
    k_create = Keyword('create')
    k_critical = Keyword('critical')
    k_database = Keyword('database')
    k_dbname = Keyword('dbname')
    k_dbpath = Keyword('dbpath')
    k_debug = Keyword('debug')
    k_derivative = Keyword('derivative')
    k_difference = Keyword('difference')
    k_drop = Keyword('drop')
    k_drop_threshold = Keyword('drop_threshold')
    k_duration_log = Keyword('duration_log')
    k_duration_num = Keyword('duration_num')
    k_end = Keyword('end')
    k_error = Keyword('error')
    k_expiration_log = Keyword('expiration_log')
    k_expiration_num = Keyword('expiration_num')
    k_expression = Keyword('expression')
    k_false = Keyword('false')
    k_fifo_files = Keyword('fifo_files')
    k_filter = Keyword('filter')
    k_first = Keyword('first')
    k_float = Keyword('float')
    k_for = Keyword('for')
    k_from = Keyword('from')
    k_full = Keyword('full')
    k_grant = Keyword('grant')
    k_group = Keyword('group')
    k_groups = Keyword('groups')
    k_help = Choice(Keyword('help'), Token('?'))
    k_idle_percentage = Keyword('idle_percentage')
    k_idle_time = Keyword('idle_time')
    k_inf = Keyword('inf')
    k_info = Keyword('info')
    k_ignore_threshold = Keyword('ignore_threshold')
    k_insert = Keyword('insert')
    k_integer = Keyword('integer')
    k_intersection = Choice(
        Token('&'),
        Keyword('intersection'),
        most_greedy=False)
    k_interval = Keyword('interval')
    k_ip_support = Keyword('ip_support')
    k_last = Keyword('last')
    k_length = Keyword('length')
    k_libuv = Keyword('libuv')
    k_limit = Keyword('limit')
    k_list = Keyword('list')
    k_list_limit = Keyword('list_limit')
    k_log = Keyword('log')
    k_log_level = Keyword('log_level')
    k_max = Keyword('max')
    k_max_open_files = Keyword('max_open_files')
    k_mean = Keyword('mean')
    k_median = Keyword('median')
    k_median_high = Keyword('median_high')
    k_median_low = Keyword('median_low')
    k_mem_usage = Keyword('mem_usage')
    k_merge = Keyword('merge')
    k_min = Keyword('min')
    k_modify = Keyword('modify')
    k_name = Keyword('name')
    k_nan = Keyword('nan')
    k_ninf = Sequence('-', k_inf)
    k_now = Keyword('now')
    k_number = Keyword('number')
    k_online = Keyword('online')
    k_open_files = Keyword('open_files')
    k_or = Keyword('or')
    k_password = Keyword('password')
    k_points = Keyword('points')
    k_pool = Keyword('pool')
    k_pools = Keyword('pools')
    k_port = Keyword('port')
    k_prefix = Keyword('prefix')
    k_pvariance = Keyword('pvariance')
    k_read = Keyword('read')
    k_received_points = Keyword('received_points')
    k_reindex_progress = Keyword('reindex_progress')
    k_revoke = Keyword('revoke')
    k_select = Keyword('select')
    k_select_points_limit = Keyword('select_points_limit')
    k_selected_points = Keyword('selected_points')
    k_series = Keyword('series')
    k_server = Keyword('server')
    k_servers = Keyword('servers')
    k_set = Keyword('set')
    k_shard_duration = Keyword('shard_duration')
    k_shards = Keyword('shards')
    k_show = Keyword('show')
    k_sid = Keyword('sid')
    k_size = Keyword('size')
    k_start = Keyword('start')
    k_startup_time = Keyword('startup_time')
    k_status = Keyword('status')
    k_stddev = Keyword('stddev')
    k_string = Keyword('string')
    k_suffix = Keyword('suffix')
    k_sum = Keyword('sum')
    k_symmetric_difference = Choice(
        Token('^'),
        Keyword('symmetric_difference'),
        most_greedy=False)
    k_sync_progress = Keyword('sync_progress')
    k_tag = Keyword('tag')
    k_tags = Keyword('tags')
    k_tee_pipe_name = Keyword('tee_pipe_name')
    k_time_precision = Keyword('time_precision')
    k_timeit = Keyword('timeit')
    k_timeval = Keyword('timeval')
    k_timezone = Keyword('timezone')
    k_to = Keyword('to')
    k_true = Keyword('true')
    k_type = Keyword('type')
    k_union = Choice(
        Tokens(', |'),
        Keyword('union'),
        most_greedy=False)
    k_untag = Keyword('untag')
    k_uptime = Keyword('uptime')
    k_user = Keyword('user')
    k_users = Keyword('users')
    k_using = Keyword('using')
    k_uuid = Keyword('uuid')
    k_variance = Keyword('variance')
    k_version = Keyword('version')
    k_warning = Keyword('warning')
    k_where = Keyword('where')
    k_who_am_i = Keyword('who_am_i')
    k_write = Keyword('write')
    c_difference = Choice(
        Token('-'),
        k_difference,
        most_greedy=False)

    access_keywords = Choice(
        k_read,
        k_write,
        k_modify,
        k_full,
        k_select,
        k_show,
        k_list,
        k_count,
        k_create,
        k_insert,
        k_drop,
        k_grant,
        k_revoke,
        k_alter,
        most_greedy=False)

    _boolean = Choice(k_true, k_false, most_greedy=False)

    log_keywords = Choice(
        k_debug,
        k_info,
        k_warning,
        k_error,
        k_critical,
        most_greedy=False)

    int_expr = Prio(
        r_integer,
        Sequence('(', THIS, ')'),
        Sequence(THIS, Tokens('+ - * % /'), THIS))

    string = Choice(r_singleq_str, r_doubleq_str, most_greedy=False)

    time_expr = Prio(
        r_time_str,
        k_now,
        string,
        r_integer,
        Sequence('(', THIS, ')'),
        Sequence(THIS, Tokens('+ - * % /'), THIS))

    series_columns = List(Choice(
        k_name,
        k_type,
        k_length,
        k_start,
        k_end,
        k_shard_duration,
        k_pool,
        most_greedy=False), ',', 1)

    shard_columns = List(Choice(
        k_sid,
        k_pool,
        k_server,
        k_size,
        k_start,
        k_end,
        k_type,
        k_status,
        most_greedy=False), ',', 1)

    server_columns = List(Choice(
        # Local properties
        k_address,
        k_buffer_path,
        k_buffer_size,
        k_dbpath,
        k_ip_support,
        k_libuv,
        k_name,
        k_port,
        k_uuid,
        k_pool,
        k_version,
        k_online,
        k_startup_time,
        k_status,
        # Remote properties
        k_active_handles,
        k_active_tasks,
        k_fifo_files,
        k_idle_percentage,
        k_idle_time,
        k_log_level,
        k_max_open_files,
        k_mem_usage,
        k_open_files,
        k_received_points,
        k_reindex_progress,
        k_selected_points,
        k_sync_progress,
        k_tee_pipe_name,
        k_uptime,
        most_greedy=False), ',', 1)

    group_columns = List(Choice(
        k_expression,
        k_name,
        k_series,
        most_greedy=False), ',', 1)

    user_columns = List(Choice(
        k_name,
        k_access,
        most_greedy=False), ',', 1)

    tag_columns = List(Choice(
        k_name,
        k_series,
        most_greedy=False), ',', 1)

    pool_props = Choice(
        k_pool,
        k_servers,
        k_series,
        most_greedy=False)
    pool_columns = List(pool_props, ',', 1)

    bool_operator = Tokens('== !=')
    int_operator = Tokens('< > == != <= >=')
    str_operator = Tokens('< > == != <= >= ~ !~')

    # where group
    where_group = Sequence(k_where, Prio(
        Sequence(k_series, int_operator, int_expr),
        Sequence(
            Choice(k_expression, k_name, most_greedy=False),
            str_operator,
            string),
        Sequence('(', THIS, ')'),
        Sequence(THIS, k_and, THIS),
        Sequence(THIS, k_or, THIS)))

    # where tag
    where_tag = Sequence(k_where, Prio(
        Sequence(k_name, str_operator, string),
        Sequence(k_series, int_operator, int_expr),
        Sequence('(', THIS, ')'),
        Sequence(THIS, k_and, THIS),
        Sequence(THIS, k_or, THIS)))

    # where pool
    where_pool = Sequence(k_where, Prio(
        Sequence(pool_props, int_operator, int_expr),
        Sequence('(', THIS, ')'),
        Sequence(THIS, k_and, THIS),
        Sequence(THIS, k_or, THIS)))

    # where series
    where_series = Sequence(k_where, Prio(
        Sequence(
            Choice(k_length, k_pool, most_greedy=False),
            int_operator,
            int_expr),
        Sequence(k_name, str_operator, string),
        Sequence(
            Choice(k_start, k_end, k_shard_duration, most_greedy=False),
            int_operator,
            time_expr),
        Sequence(
            k_type,
            bool_operator,
            Choice(k_string, k_integer, k_float, most_greedy=False)),
        Sequence('(', THIS, ')'),
        Sequence(THIS, k_and, THIS),
        Sequence(THIS, k_or, THIS)))

    # where server
    where_server = Sequence(k_where, Prio(
        Sequence(Choice(
            k_active_handles,
            k_active_tasks,
            k_buffer_size,
            k_fifo_files,
            k_idle_percentage,
            k_idle_time,
            k_port,
            k_pool,
            k_startup_time,
            k_max_open_files,
            k_mem_usage,
            k_open_files,
            k_received_points,
            k_selected_points,
            k_uptime,
            most_greedy=False), int_operator, int_expr),
        Sequence(Choice(
            k_address,
            k_buffer_path,
            k_dbpath,
            k_ip_support,
            k_libuv,
            k_name,
            k_uuid,
            k_version,
            k_status,
            k_reindex_progress,
            k_sync_progress,
            k_tee_pipe_name,
            most_greedy=False), str_operator, string),
        Sequence(k_online, bool_operator, _boolean),
        Sequence(k_log_level, int_operator, log_keywords),
        Sequence('(', THIS, ')'),
        Sequence(THIS, k_and, THIS),
        Sequence(THIS, k_or, THIS)))

    # where shard
    where_shard = Sequence(k_where, Prio(
        Sequence(
            Choice(k_sid, k_pool, k_size, most_greedy=False),
            int_operator,
            int_expr),
        Sequence(Choice(k_server, k_status), str_operator, string),
        Sequence(
            Choice(k_start, k_end, most_greedy=False),
            int_operator,
            time_expr),
        Sequence(
            k_type,
            bool_operator,
            Choice(k_number, k_log, most_greedy=False)),
        Sequence('(', THIS, ')'),
        Sequence(THIS, k_and, THIS),
        Sequence(THIS, k_or, THIS)))

    # where user
    where_user = Sequence(k_where, Prio(
        Sequence(k_name, str_operator, string),
        Sequence(k_access, int_operator, access_keywords),
        Sequence('(', THIS, ')'),
        Sequence(THIS, k_and, THIS),
        Sequence(THIS, k_or, THIS)))

    series_setopr = Choice(
        k_union,
        c_difference,
        k_intersection,
        k_symmetric_difference,
        most_greedy=False)

    series_parentheses = Sequence('(', THIS, ')')

    series_all = Choice(Token('*'), k_all, most_greedy=False)
    series_name = Repeat(string, 1, 1)
    group_name = Repeat(r_grave_str, 1, 1)
    tag_name = Repeat(r_grave_str, 1, 1)
    series_re = Repeat(r_regex, 1, 1)
    uuid = Choice(r_uuid_str, string, most_greedy=False)
    group_tag_match = Repeat(r_grave_str, 1, 1)
    series_match = Prio(
        List(Choice(
            series_all,
            series_name,
            group_tag_match,
            series_re,
            most_greedy=False), series_setopr, 1),
        Choice(
            series_all,
            series_name,
            group_tag_match,
            series_re,
            most_greedy=False),
        series_parentheses,
        Sequence(THIS, series_setopr, THIS),
    )

    limit_expr = Sequence(k_limit, int_expr)

    before_expr = Sequence(k_before, time_expr)
    after_expr = Sequence(k_after, time_expr)
    between_expr = Sequence(k_between, time_expr, k_and, time_expr)
    access_expr = List(access_keywords, ',', 1)

    prefix_expr = Sequence(k_prefix, string)
    suffix_expr = Sequence(k_suffix, string)

    f_all = Choice(Token('*'), k_all, most_greedy=False)

    f_points = Repeat(k_points, 1, 1)  # DEPRECATED

    f_difference = Sequence(
        k_difference,
        '(',
        Optional(time_expr),
        ')')
    f_derivative = Sequence(
        k_derivative,
        '(',
        List(time_expr, ',', 0, 2),
        ')')
    f_mean = Sequence(
        k_mean,
        '(', Optional(time_expr), ')')
    f_median = Sequence(
        k_median,
        '(', Optional(time_expr), ')')
    f_median_low = Sequence(
        k_median_low,
        '(', Optional(time_expr), ')')
    f_median_high = Sequence(
        k_median_high,
        '(', Optional(time_expr), ')')
    f_sum = Sequence(
        k_sum,
        '(', Optional(time_expr), ')')
    f_min = Sequence(
        k_min,
        '(', Optional(time_expr), ')')
    f_max = Sequence(
        k_max,
        '(', Optional(time_expr), ')')
    f_count = Sequence(
        k_count,
        '(', Optional(time_expr), ')')
    f_variance = Sequence(
        k_variance,
        '(', Optional(time_expr), ')')
    f_pvariance = Sequence(
        k_pvariance,
        '(', Optional(time_expr), ')')
    f_stddev = Sequence(
        k_stddev,
        '(', Optional(time_expr), ')')
    f_first = Sequence(
        k_first,
        '(', Optional(time_expr), ')')
    f_last = Sequence(
        k_last,
        '(', Optional(time_expr), ')')
    f_timeval = Sequence(
        k_timeval,
        '(', ')')
    f_interval = Sequence(
        k_interval,
        '(', ')')

    f_filter = Sequence(
        k_filter,
        '(',
        Optional(str_operator),
        Choice(
            string,
            r_integer,
            r_float,
            r_regex,
            k_nan,
            k_inf,
            k_ninf,
            most_greedy=True),
        ')')
    f_limit = Sequence(
        k_limit,
        '(',
        int_expr,
        ',',
        Choice(
            k_mean,
            k_median,
            k_median_high,
            k_median_low,
            k_sum,
            k_min,
            k_max,
            k_count,
            k_variance,
            k_pvariance,
            k_stddev,
            k_first,
            k_last,
            most_greedy=False),
        ')')

    aggregate_functions = List(Choice(
        f_all,
        f_limit,
        f_mean,
        f_sum,
        f_median,
        f_median_low,
        f_median_high,
        f_min,
        f_max,
        f_count,
        f_variance,
        f_pvariance,
        f_stddev,
        f_first,
        f_last,
        f_timeval,
        f_interval,
        f_difference,
        f_derivative,
        f_filter,
        f_points,
        most_greedy=False), '=>', 1)

    select_aggregate = Sequence(
        aggregate_functions,
        Optional(prefix_expr),
        Optional(suffix_expr))

    select_aggregates = List(select_aggregate, ',', 1)

    merge_as = Sequence(
        k_merge,
        k_as,
        string,
        Optional(Sequence(k_using, aggregate_functions)))

    set_address = Sequence(k_set, k_address, string)
    set_tee_pipe_name = Sequence(k_set, k_tee_pipe_name, Choice(
        k_false,
        string,
        most_greedy=False))
    set_backup_mode = Sequence(k_set, k_backup_mode, _boolean)
    set_drop_threshold = Sequence(k_set, k_drop_threshold, r_float)
    set_expression = Sequence(k_set, k_expression, r_regex)
    set_ignore_threshold = Sequence(k_set, k_ignore_threshold, _boolean)
    set_list_limit = Sequence(k_set, k_list_limit, r_uinteger)
    set_log_level = Sequence(k_set, k_log_level, log_keywords)
    set_name = Sequence(k_set, k_name, string)
    set_password = Sequence(k_set, k_password, string)
    set_port = Sequence(k_set, k_port, r_uinteger)
    set_select_points_limit = Sequence(
        k_set, k_select_points_limit, r_uinteger)
    set_timezone = Sequence(k_set, k_timezone, string)
    tag_series = Sequence(k_tag, tag_name)
    untag_series = Sequence(k_untag, tag_name)
    set_expiration_num = Sequence(
        k_set,
        k_expiration_num,
	time_expr,
        Optional(set_ignore_threshold))
    set_expiration_log = Sequence(
        k_set,
        k_expiration_log,
	time_expr,
        Optional(set_ignore_threshold))

    alter_database = Sequence(k_database, Choice(
        set_drop_threshold,
        set_list_limit,
        set_select_points_limit,
        set_timezone,
        set_expiration_num,
        set_expiration_log,
        most_greedy=False))

    alter_group = Sequence(k_group, group_name, Choice(
        set_expression,
        set_name,
        most_greedy=False))

    alter_tag = Sequence(k_tag, tag_name, Choice(
        set_name,
        most_greedy=False))

    alter_server = Sequence(k_server, uuid, Choice(
        set_log_level,
        set_backup_mode,
        set_tee_pipe_name,
        set_address,
        set_port,
        most_greedy=False))

    alter_servers = Sequence(k_servers, Optional(where_server), Choice(
        set_log_level,
        set_tee_pipe_name,
        most_greedy=False))

    alter_user = Sequence(k_user, string, Choice(
        set_password,
        set_name,
        most_greedy=False))

    alter_series = Sequence(
        k_series,
        series_match,
        Optional(where_series),
        Choice(tag_series, untag_series, most_greedy=False))

    count_groups = Sequence(
        k_groups, Optional(where_group))
    count_tags = Sequence(
        k_tags, Optional(where_tag))
    count_pools = Sequence(
        k_pools, Optional(where_pool))
    count_series = Sequence(
        k_series, Optional(series_match), Optional(where_series))
    count_servers = Sequence(
        k_servers, Optional(where_server))
    count_servers_received = Sequence(
        k_servers,
        k_received_points,
        Optional(where_server))
    count_servers_selected = Sequence(
        k_servers,
        k_selected_points,
        Optional(where_server))
    count_shards = Sequence(
        k_shards, Optional(where_shard))
    count_shards_size = Sequence(
        k_shards, k_size, Optional(where_shard))
    count_users = Sequence(
        k_users, Optional(where_user))
    count_series_length = Sequence(
        k_series,
        k_length,
        Optional(series_match),
        Optional(where_series))

    create_group = Sequence(
        k_group, group_name, k_for, r_regex)
    create_user = Sequence(
        k_user, string, set_password)

    drop_group = Sequence(k_group, group_name)
    drop_tag = Sequence(k_tag, tag_name)

    # Drop statement needs at least a series_math or where STMT or both
    drop_series = Sequence(
        k_series,
        Optional(series_match),
        Optional(where_series),
        Optional(set_ignore_threshold))
    drop_shards = Sequence(
        k_shards,
        Optional(where_shard),
        Optional(set_ignore_threshold))
    drop_server = Sequence(k_server, uuid)
    drop_user = Sequence(k_user, string)

    grant_user = Sequence(
        k_user, string, Optional(set_password))

    list_groups = Sequence(
        k_groups, Optional(group_columns), Optional(where_group))
    list_tags = Sequence(
        k_tags, Optional(tag_columns), Optional(where_tag))
    list_pools = Sequence(
        k_pools, Optional(pool_columns), Optional(where_pool))
    list_series = Sequence(
        k_series,
        Optional(series_columns),
        Optional(series_match),
        Optional(where_series))
    list_servers = Sequence(
        k_servers, Optional(server_columns), Optional(where_server))
    list_shards = Sequence(
        k_shards, Optional(shard_columns), Optional(where_shard))
    list_users = Sequence(
        k_users, Optional(user_columns), Optional(where_user))

    revoke_user = Sequence(k_user, string)

    alter_stmt = Sequence(k_alter, Choice(
        alter_series,
        alter_user,
        alter_group,
        alter_tag,
        alter_server,
        alter_servers,
        alter_database,
        most_greedy=False))

    calc_stmt = Repeat(time_expr, 1, 1)

    count_stmt = Sequence(k_count, Choice(
        count_groups,
        count_pools,
        count_series,
        count_servers,
        count_servers_received,
        count_servers_selected,
        count_shards,
        count_shards_size,
        count_users,
        count_tags,
        count_series_length,
        most_greedy=True))

    create_stmt = Sequence(k_create, Choice(
        create_group,
        create_user))

    drop_stmt = Sequence(k_drop, Choice(
        drop_group,
        drop_tag,
        drop_series,
        drop_shards,
        drop_server,
        drop_user,
        most_greedy=False))

    grant_stmt = Sequence(k_grant, access_expr, k_to, Choice(
        grant_user,
        most_greedy=False))

    list_stmt = Sequence(k_list, Choice(
        list_series,
        list_tags,
        list_users,
        list_shards,
        list_groups,
        list_servers,
        list_pools,
        most_greedy=False
    ), Optional(limit_expr))

    revoke_stmt = Sequence(k_revoke, access_expr, k_from, Choice(
        revoke_user,
        most_greedy=False))

    select_stmt = Sequence(
        k_select,
        select_aggregates,
        k_from,
        series_match,
        Optional(where_series),
        Optional(Choice(
            after_expr,
            between_expr,
            before_expr,
            most_greedy=False)),
        Optional(merge_as))

    show_stmt = Sequence(k_show, List(Choice(
        k_active_handles,
        k_active_tasks,
        k_buffer_path,
        k_buffer_size,
        k_dbname,
        k_dbpath,
        k_drop_threshold,
        k_duration_log,
        k_duration_num,
        k_fifo_files,
        k_expiration_log,
        k_expiration_num,
        k_idle_percentage,
        k_idle_time,
        k_ip_support,
        k_libuv,
        k_list_limit,
        k_log_level,
        k_max_open_files,
        k_mem_usage,
        k_open_files,
        k_pool,
        k_received_points,
        k_reindex_progress,
        k_selected_points,
        k_select_points_limit,
        k_server,
        k_startup_time,
        k_status,
        k_sync_progress,
        k_tee_pipe_name,
        k_time_precision,
        k_timezone,
        k_uptime,
        k_uuid,
        k_version,
        k_who_am_i,
        most_greedy=False), ',', 0))

    timeit_stmt = Repeat(k_timeit, 1, 1)

    help_stmt = Ref()

    START = Sequence(
        Optional(timeit_stmt),
        Optional(Choice(
            select_stmt,
            list_stmt,
            count_stmt,
            alter_stmt,
            create_stmt,
            drop_stmt,
            grant_stmt,
            revoke_stmt,
            show_stmt,
            calc_stmt,
            help_stmt,
            most_greedy=False)),
        Optional(r_comment))

    help_stmt = Sequence(k_help)  # Dummy


def _set_attribute(cls, name, value):
    setattr(cls, name, value)
    if name not in cls._order:
        cls._order.append(name)
    if hasattr(value, 'name'):
        raise SyntaxError('Element name is set to {0!r} and therefore cannot '
                          'be set to {1!r}. Use Repeat({0}, 1, 1) as a '
                          'workaround.'.format(value.name, name))
    value.name = name


def _walk(cls, path, structure):
    name = '_'.join(path)
    try:
        if structure:
            for child_path, child in structure.items():
                _walk(cls, child_path, child)

            opt = Optional(Choice(*[
                getattr(cls, '_'.join(p)) for p in structure.keys()]))

            value = Sequence(*[getattr(
                cls,
                'k_' + path[-1],
                Keyword(path[-1])), opt])
        else:
            value = Keyword(path[-1])
    except AttributeError:
        logging.critical('Cannot parse help file: {!r}'.format(
            '{}.md'.format('_'.join(path))))
        import sys
        sys.exit(1)
    else:
        if name != 'help':
            _set_attribute(cls, name, value)
        else:
            # Replace the ref element
            cls.help_stmt._element = value


def _build_help(cls):
    '''Add 'help' statements to this class.

    Help files are MarkDown files which are read by the help module.
    '''
    for path, structure in help_structure.items():
        _walk(cls, path, structure)

_build_help(SiriGrammar)

siri_grammar = SiriGrammar()



import os
import asyncio
import functools
import random
import time
import math
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import SiriDB
from testing import TestBase
from testing import SiriDBAsyncUnixServer
from testing import parse_args


PIPE_NAME = '/tmp/dbtest_tee.sock'

DATA = {
    'series float': [
        [1471254705, 1.5],
        [1471254707, -3.5],
        [1471254710, -7.3]],
    'series integer': [
        [1471254705, 5],
        [1471254708, -3],
        [1471254710, -7]],
    'aggr': [
        [1447249033, 531], [1447249337, 534],
        [1447249633, 535], [1447249937, 531],
        [1447250249, 532], [1447250549, 537],
        [1447250868, 530], [1447251168, 520],
        [1447251449, 54], [1447251749, 54],
        [1447252049, 513], [1447252349, 537],
        [1447252649, 528], [1447252968, 531],
        [1447253244, 533], [1447253549, 538],
        [1447253849, 534], [1447254149, 532],
        [1447254449, 533], [1447254748, 537]],
    'huge': [
        [1471254705, 9223372036854775807],
        [1471254706, 9223372036854775806],
        [1471254707, 9223372036854775805],
        [1471254708, 9223372036854775804]],
    'equal ts': [
        [1471254705, 0], [1471254705, 1], [1471254705, 1],
        [1471254707, 0], [1471254707, 1], [1471254708, 0],
    ],
    'variance': [
        [1471254705, 2.75], [1471254706, 1.75], [1471254707, 1.25],
        [1471254708, 0.25], [1471254709, 0.5], [1471254710, 1.25],
        [1471254711, 3.5]
    ],
    'pvariance': [
        [1471254705, 0.0], [1471254706, 0.25], [1471254707, 0.25],
        [1471254708, 1.25], [1471254709, 1.5], [1471254710, 1.75],
        [1471254711, 2.75], [1471254712, 3.25]
    ],
    'filter': [
        [1471254705, 5],
        [1471254710, -3],
        [1471254715, -7],
        [1471254720, 7]
    ],
    'one': [
        [1471254710, 1]
    ],
    'log': [
        [1471254710, 'log line one'],
        [1471254712, 'log line two'],
        [1471254714, 'another line (three)'],
        [1471254716, 'and yet one more'],
    ],
    # 'special': [
    #     [1471254705, 0.1],
    #     [1471254706, math.nan],
    #     [1471254707, math.inf],
    #     [1471254708, -math.inf],
    # ]
}

if os.path.exists(PIPE_NAME):
    os.unlink(PIPE_NAME)


class TestTee(TestBase):
    title = 'Test tee'

    def on_data(self, data):
        for k, v in data.items():
            if k not in self._tee_data:
                self._tee_data[k] = []
            self._tee_data[k].extend(v)

    @default_test_setup(2)
    async def run(self):
        self._tee_data = {}

        server = SiriDBAsyncUnixServer(PIPE_NAME, self.on_data)

        await server.create()

        await self.client0.connect()

        await asyncio.sleep(1)

        await self.client0.query(
            'alter servers set tee_pipe_name "{}"'.format(PIPE_NAME))

        await asyncio.sleep(1)

        self.assertEqual(
            await self.client0.insert(DATA),
            {'success_msg': 'Successfully inserted 60 point(s).'})

        self.assertAlmostEqual(
            await self.client0.query('select * from "series float"'),
            {'series float': DATA['series float']})

        self.assertEqual(
            await self.client0.query('select * from "series integer"'),
            {'series integer': DATA['series integer']})

        self.assertEqual(
            await self.client0.query('select * from "log"'),
            {'log': DATA['log']})

        await asyncio.sleep(1)

        self.assertEqual(DATA, self._tee_data)

        await self.db.add_pool(self.server1, sleep=3)
        await self.assertIsRunning(self.db, self.client0, timeout=50)

        await self.client1.connect()

        self._tee_data = {}
        await self.client0.query('drop series set ignore_threshold true')

        await asyncio.sleep(1)

        await self.client0.query(
            'alter servers set tee_pipe_name "{}"'.format(PIPE_NAME))

        await asyncio.sleep(1)

        self.assertEqual(
            await self.client0.insert(DATA),
            {'success_msg': 'Successfully inserted 60 point(s).'})

        await asyncio.sleep(1)

        self.assertEqual(DATA, self._tee_data)

        self.client0.close()
        self.client1.close()


if __name__ == '__main__':
    parse_args()
    run_test(TestTee())



import os
import logging
import shutil
import time
import random
import functools
import string
from .constants import TEST_DIR
from .series import Series

_MAP_TS = {
    's': 10**0,
    'ms': 10**3,
    'us': 10**6,
    'ns': 10**9
}


def cleanup():
    logging.info('Remove test dir')
    try:
        shutil.rmtree(os.path.join(TEST_DIR))
    except FileNotFoundError:
        pass
    os.mkdir(TEST_DIR)

    logging.info('Force kill all open siridb-server processes')
    os.system('pkill -9 siridb-server')

    logging.info('Force kill all open memcheck-amd64- processes')
    os.system('pkill -9 memcheck-amd64-')


def random_value(tp=float, mi=-100, ma=100):
    i = random.randrange(mi, ma)
    if tp == float:
        return i * random.random()
    elif tp == int:
        return i


def random_series_name(size=12, chars=string.ascii_letters + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


def gen_points(
        n=100,
        time_precision='s',
        tp=float,
        mi=-100,
        ma=100,
        ts_gap=1,
        shuffle=True):
    if isinstance(ts_gap, str):
        if ts_gap.endswith('s'):
            ts_gap = int(ts_gap[:-1]) * _MAP_TS[time_precision]
        elif ts_gap.endswith('m'):
            ts_gap = int(ts_gap[:-1]) * _MAP_TS[time_precision] * 60
        elif ts_gap.endswith('h'):
            ts_gap = int(ts_gap[:-1]) * _MAP_TS[time_precision] * 3600
        elif ts_gap.endswith('d'):
            ts_gap = int(ts_gap[:-1]) * _MAP_TS[time_precision] * 3600 * 24
        elif ts_gap.endswith('w'):
            ts_gap = int(ts_gap[:-1]) * _MAP_TS[time_precision] * 3600 * 24 * 7

    end = int(time.time() * _MAP_TS[time_precision])
    start = end - (n * ts_gap)
    timestamps = list(range(start, end, ts_gap))
    if shuffle:
        random.shuffle(timestamps)
    return [[ts, random_value(tp, mi, ma)] for ts in timestamps]


def gen_data(points=functools.partial(gen_points), n=100):
    return {random_series_name(): points() for _ in range(n)}


def gen_series(n=10000):
    return [Series(random_series_name()) for _ in range(n)]



import os
import asyncio
import functools
import random
import time
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import SiriDBAsyncUnixConnection
from testing import parse_args


PIPE_NAME = '/tmp/siridb_pipe_test.sock'

DATA = {
    'series num_float': [
        [1471254705, 1.5],
        [1471254707, -3.5],
        [1471254710, -7.3]],
    'series num_integer': [
        [1471254705, 5],
        [1471254708, -3],
        [1471254710, -7]],
    'series_log': [
        [1471254710, 'log line one'],
        [1471254712, 'log line two'],
        [1471254714, 'another line (three)'],
        [1471254716, 'and yet one more']]
}

if os.path.exists(PIPE_NAME):
    os.unlink(PIPE_NAME)


class TestPipeSupport(TestBase):
    title = 'Test pipe support object'

    @default_test_setup(1, pipe_name=PIPE_NAME)
    async def run(self):

        pipe_client0 = SiriDBAsyncUnixConnection(PIPE_NAME)

        await pipe_client0.connect('iris', 'siri', self.db.dbname)

        self.assertEqual(
            await pipe_client0.insert(DATA),
            {'success_msg': 'Successfully inserted 10 point(s).'})

        self.assertAlmostEqual(
            await pipe_client0.query('select * from "series num_float"'),
            {'series num_float': DATA['series num_float']})

        self.assertEqual(
            await pipe_client0.query('select * from "series num_integer"'),
            {'series num_integer': DATA['series num_integer']})

        self.assertEqual(
            await pipe_client0.query('select * from "series_log"'),
            {'series_log': DATA['series_log']})

        pipe_client0.close()


if __name__ == '__main__':
    parse_args()
    run_test(TestPipeSupport())



import asyncio
import functools
import random
import time
import re
import calendar
import datetime
import collections
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import parse_args


# Compression OFF:
# du --bytes testdir/dbpath0/dbtest/shards/
# 423114  testdir/dbpath0/dbtest/shards/
# 416266  (optimized)

# Compression ON:
# du --bytes testdir/dbpath0/dbtest/shards/
# 222314  testdir/dbpath0/dbtest/shards/
# 153380  (optimized)

SYSLOG = '/home/joente/syslog.log'
FMT = '%b %d %H:%M:%S'
MTCH = re.compile(
    r'(\w\w\w\s[\d\s]\d\s\d\d:\d\d:'
    r'\d\d)\s(\w+)\s([\w\-\.\/@]+)(\[\d+\])?:\s(.*)')


class TestSyslog(TestBase):
    title = 'Test with syslog data'

    async def insert_syslog(self, batch_size=100):

        with open(SYSLOG, 'r') as f:
            lines = f.readlines()

        points = collections.defaultdict(list)
        n = 0

        for line in lines:
            r = MTCH.match(line.strip())
            if not r:
                continue

            rtime, host, process, pid, logline = r.groups()
            dt = datetime.datetime.strptime(rtime, FMT)
            dt = dt.replace(year=datetime.datetime.now().year)
            ts = calendar.timegm(dt.timetuple())
            points['{}|{}'.format(host, process)].append([ts, logline])
            n += 1
            if n % batch_size == 0:
                await self.client0.insert(points)
                points.clear()

        if points:
            await self.client0.insert(points)

    @default_test_setup(2, compression=True, duration_log='1w')
    async def run(self):
        await self.client0.connect()

        # await self.db.add_pool(self.server1, sleep=30)

        await self.db.add_replica(self.server1, 0, sleep=30)

        await self.insert_syslog()

        await self.client0.query('select * from /.*vbox.*/ merge as "t"')

        self.client0.close()


if __name__ == '__main__':
    parse_args()
    run_test(TestSyslog())



class Series:
    def __init__(self, name):
        self.name = name
        self.points = []
        self._points = None

    def add_points(self, points):
        self.commit_points()
        self._points = points
        return points

    def commit_points(self):
        if self._points:
            for point in self._points:
                self.points.append(point)
            self._points = None



import os
import logging
import configparser
import subprocess
import psutil
import asyncio
import time
import socket
import platform
from .constants import SIRIDBC
from .constants import TEST_DIR
from .constants import VALGRIND
from .constants import MAX_OPEN_FILES
from .color import Color


MEM_PROC = \
    'memcheck-amd64-' if platform.architecture()[0] == '64bit' else \
    'memcheck-x86-li'


def get_file_content(fn):
    with open(fn, 'r') as f:
        data = f.read()
        return data


class Server:
    HOLD_TERM = False
    GEOMETRY = '140x60'
    MEM_CHECK = False
    BUILDTYPE = 'Release'
    SERVER_ADDRESS = '%HOSTNAME'
    IP_SUPPORT = 'ALL'  # ALL, IPV4ONLY
    TERMINAL = None  # one of [ 'XTERM', 'XFCE4_TERMINAL', None  ]
    BIND_CLIENT_ADDRESS = "::"
    BIND_SERVER_ADDRESS = "::"

    def __init__(self,
                 n,
                 title,
                 optimize_interval=300,
                 heartbeat_interval=30,
                 buffer_sync_interval=500,
                 compression=True,
                 auto_duration=True,
                 pipe_name=None,
                 **unused):
        self.n = n
        self.test_title = title.lower().replace(' ', '_')
        self.compression = compression
        self.auto_duration = auto_duration
        self.enable_pipe_support = int(bool(pipe_name))
        self.pipe_name = \
            'siridb_client.sock' if not self.enable_pipe_support else \
            pipe_name
        self.http_status_port = 8080 + n
        self.listen_client_port = 9000 + n
        self.listen_backend_port = 9010 + n
        self.http_api_port = 9020 + n
        self.buffer_sync_interval = buffer_sync_interval
        self._server_address = self.SERVER_ADDRESS
        self.server_address = \
            self._server_address.lstrip('[').rstrip(']').replace(
                '%HOSTNAME', socket.gethostname())
        self.ip_support = self.IP_SUPPORT
        self.optimize_interval = optimize_interval
        self.heartbeat_interval = heartbeat_interval
        self.bind_client_address = self.BIND_CLIENT_ADDRESS
        self.bind_server_address = self.BIND_SERVER_ADDRESS
        self.cfgfile = os.path.join(TEST_DIR, 'siridb{}.conf'.format(self.n))
        self.dbpath = os.path.join(TEST_DIR, 'dbpath{}'.format(self.n))
        self.name = 'SiriDB:{}'.format(self.listen_backend_port)
        self.pid = None
        self.buffer_path = None
        self.buffer_size = None

    @property
    def addr(self):
        return '{}:{}'.format(self.server_address, self.listen_client_port)

    def create(self):
        logging.info('Create server {}'.format(self.name))

        config = configparser.RawConfigParser()
        config.add_section('siridb')
        config.set('siridb', 'listen_client_port', self.listen_client_port)
        config.set('siridb', 'bind_client_address', self.bind_client_address)
        config.set('siridb', 'bind_server_address', self.bind_server_address)
        config.set('siridb', 'server_name', '{}:{}'.format(
            self._server_address,
            self.listen_backend_port))
        config.set('siridb', 'ip_support', self.ip_support)
        config.set('siridb', 'optimize_interval', self.optimize_interval)
        config.set('siridb', 'heartbeat_interval', self.heartbeat_interval)
        config.set('siridb', 'buffer_sync_interval', self.buffer_sync_interval)
        config.set('siridb', 'default_db_path', self.dbpath)
        config.set('siridb', 'max_open_files', MAX_OPEN_FILES)
        config.set('siridb', 'enable_shard_compression', int(self.compression))
        config.set(
            'siridb', 'enable_shard_auto_duration',
            int(self.auto_duration))
        config.set('siridb', 'enable_pipe_support', self.enable_pipe_support)
        config.set('siridb', 'pipe_client_name',  self.pipe_name)
        config.set('siridb', 'http_status_port',  self.http_status_port)
        config.set('siridb', 'http_api_port',  self.http_api_port)

        with open(self.cfgfile, 'w') as configfile:
            config.write(configfile)

        try:
            os.mkdir(self.dbpath)
        except FileExistsError:
            pass

    def _get_pid_set(self):
        try:
            ret = set(map(int, subprocess.check_output([
                'pgrep',
                MEM_PROC if self.MEM_CHECK else 'siridb-server']).split()))
        except subprocess.CalledProcessError:
            ret = set()
        return ret

    async def start(self, sleep=None):
        prev = self._get_pid_set()
        if self.TERMINAL == 'xfce4-terminal':
            self.proc = subprocess.Popen(
                'xfce4-terminal -e "{}{} --config {} --log-colorized"'
                ' --title {} --geometry={}{}'
                .format(VALGRIND if self.MEM_CHECK else '',
                        SIRIDBC.format(BUILDTYPE=self.BUILDTYPE),
                        self.cfgfile,
                        self.name,
                        self.GEOMETRY,
                        ' -H' if self.HOLD_TERM else ''),
                shell=True)
        elif self.TERMINAL == 'xterm':
            self.proc = subprocess.Popen(
               'xterm {}-title {} -geometry {} -e "{}{} --config {}"'
               .format('-hold ' if self.HOLD_TERM else '',
                       self.name,
                       self.GEOMETRY,
                       VALGRIND if self.MEM_CHECK else '',
                       SIRIDBC.format(BUILDTYPE=self.BUILDTYPE),
                       self.cfgfile),
               shell=True)
        elif self.TERMINAL is None:
            errfn = f'testdir/{self.test_title}-{self.name}-err.log'
            outfn = f'testdir/{self.test_title}-{self.name}-out.log'
            with open(errfn, 'a') as err:
                with open(outfn, 'a') as out:
                    self.proc = subprocess.Popen(
                        '{}{} --config {}'
                        .format(VALGRIND if self.MEM_CHECK else '',
                                SIRIDBC.format(BUILDTYPE=self.BUILDTYPE),
                                self.cfgfile),
                        stderr=err,
                        stdout=out,
                        shell=True)

        await asyncio.sleep(5)

        my_pid = self._get_pid_set() - prev
        if len(my_pid) != 1:
            if self.TERMINAL is None:
                if os.path.exists(outfn) and os.path.getsize(outfn):
                    reasoninfo = get_file_content(outfn)
                elif os.path.exists(errfn):
                    reasoninfo = get_file_content(errfn)
                else:
                    reasoninfo = 'unknown'
                assert 0, (
                    f'{Color.error("Failed to start SiriDB server")}\n'
                    f'{Color.info(reasoninfo)}\n')
            else:
                assert 0, (
                    'Failed to start SiriDB server. A possible reason could '
                    'be that another process is using the same port.')

        self.pid = my_pid.pop()
        if sleep:
            await asyncio.sleep(sleep)

    async def stop(self, timeout=20):
        if self.is_active():
            os.system('kill {}'.format(self.pid))

            while (timeout and self.is_active()):
                await asyncio.sleep(1.0)
                timeout -= 1

        self.proc.communicate()
        assert (self.proc.returncode == 0)

        if timeout:
            self.pid = None
            return True

        return False

    async def stopstop(self):
        if self.is_active():
            os.system('kill {}'.format(self.pid))
            await asyncio.sleep(0.2)

            if self.is_active():
                os.system('kill {}'.format(self.pid))
                await asyncio.sleep(1.0)

                if self.is_active():
                    return False

        self.pid = None
        return True

    def set_buffer_size(self, db, buffer_size):
        self.buffer_size = buffer_size
        config = configparser.RawConfigParser()
        config.add_section('buffer')
        if self.buffer_path is not None:
            config.set('buffer', 'path', self.buffer_path)
        config.set('buffer', 'size', self.buffer_size)
        with open(os.path.join(
                self.dbpath, db.dbname, 'database.conf'), 'w') as f:
            config.write(f)

    def set_buffer_path(self, db, buffer_path):
        assert(buffer_path.endswith('/'))
        curfile = os.path.join(self.dbpath, db.dbname, 'buffer.dat') \
            if self.buffer_path is None else \
            os.path.join(self.buffer_path, 'buffer.dat')
        if not os.path.exists(buffer_path):
            os.makedirs(buffer_path)
        os.rename(curfile, os.path.join(buffer_path, 'buffer.dat'))
        self.buffer_path = buffer_path
        config = configparser.RawConfigParser()
        config.add_section('buffer')
        config.set('buffer', 'path', self.buffer_path)
        if self.buffer_size is not None:
            config.set('buffer', 'size', self.buffer_size)
        with open(os.path.join(
                self.dbpath, db.dbname, 'database.conf'), 'w') as f:
            config.write(f)

    def kill(self):
        print("!!!!!!!!!!!! KILLL !!!!!!!!!!")
        os.system('kill -9 {}'.format(self.pid))
        self.pid = None

    def is_active(self):
        return False if self.pid is None else psutil.pid_exists(self.pid)



import asyncio
import functools
import random
import time
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import parse_args


TIME_PRECISION = 'ms'


class TestCompression(TestBase):
    title = 'Test compression'

    GEN_POINTS = functools.partial(
        gen_points, n=100, time_precision=TIME_PRECISION)

    async def _test_series(self, client):

        result = await client.query('select * from "series float"')
        self.assertEqual(result['series float'], self.series_float)

        result = await client.query('select * from "series int"')
        self.assertEqual(result['series int'], self.series_int)

        result = await client.query(
            'list series name, length, type, start, end')
        result['series'].sort()
        self.assertEqual(
            result,
            {'columns': ['name', 'length', 'type', 'start', 'end'],
                'series': [[
                    'series float',
                    10000, 'float',
                    self.series_float[0][0],
                    self.series_float[-1][0]], [
                        'series int',
                        10000, 'integer',
                        self.series_int[0][0],
                        self.series_int[-1][0]],
                ]})

    @default_test_setup(
        1,
        time_precision=TIME_PRECISION,
        optimize_interval=500,
        compression=True)
    async def run(self):
        await self.client0.connect()

        self.series_float = gen_points(
            tp=float, n=10000, time_precision=TIME_PRECISION, ts_gap='5m')
        random.shuffle(self.series_float)
        self.series_int = gen_points(
            tp=int, n=10000, time_precision=TIME_PRECISION, ts_gap='5m')
        random.shuffle(self.series_int)

        self.assertEqual(
            await self.client0.insert({
                'series float': self.series_float,
                'series int': self.series_int
            }), {'success_msg': 'Successfully inserted 20000 point(s).'})

        self.series_float.sort()
        self.series_int.sort()

        await self._test_series(self.client0)

        await self.client0.query('drop series /.*/ set ignore_threshold true')

        # Create some random series and start 25 insert task parallel
        series = gen_series(n=40)

        for i in range(40):
            await self.client0.insert_some_series(
                    series,
                    n=0.8,
                    timeout=0,
                    points=self.GEN_POINTS)

        # Check the result
        await self.assertSeries(self.client0, series)

        for i in range(40):
            await self.client0.insert_some_series(
                    series,
                    n=0.8,
                    timeout=0,
                    points=self.GEN_POINTS)

        # Check the result
        await self.assertSeries(self.client0, series)

        self.client0.close()

        result = await self.server0.stop()
        self.assertTrue(result)

        await self.server0.start(sleep=20)
        await self.client0.connect()

        # Check the result after rebooting the server
        await self.assertSeries(self.client0, series)


if __name__ == '__main__':
    random.seed(1)
    parse_args()
    run_test(TestCompression())



import sys
import asyncio
import functools
import logging
import random
import time
from siridb.connector import SiriDBClient
from siridb.connector.lib.exceptions import AuthenticationError
from siridb.connector.lib.exceptions import InsertError
from siridb.connector.lib.exceptions import PoolError
from siridb.connector.lib.exceptions import QueryError
from siridb.connector.lib.exceptions import ServerError
from siridb.connector.lib.exceptions import UserAuthError
from .helpers import gen_points
from .server import Server


class Client:
    def __init__(self, db, servers, username='iris', password='siri'):
        self.db = db

        if isinstance(servers, Server):
            servers = [servers]

        self.cluster = SiriDBClient(
            username=username,
            password=password,
            dbname=db.dbname,
            hostlist=[
                (server.server_address, server.listen_client_port)
                for server in servers
            ])

        self.query = self.cluster.query
        self.insert = self.cluster.insert

    async def connect(self):
        logging.info('Create client connection to database {}'.format(
            self.db.dbname))
        await self.cluster.connect()

    def close(self):
        logging.info('Closing connection to database {}'.format(
            self.db.dbname))
        self.cluster.close()

    async def insert_some_series(self,
                                 series,
                                 n=0.01,
                                 timeout=None,
                                 points=functools.partial(gen_points, n=1)):
        random.shuffle(series)

        n = int(len(series) * n)

        assert (n <= len(series) and n > 0)

        data = {s.name: s.add_points(points()) for s in series[:n]}

        while True:
            try:
                await self.insert(data)
            except PoolError as e:
                if not timeout:
                    raise e
                timeout -= 1
                await asyncio.sleep(1.0)
            else:
                break

        if timeout is not None:
            time.sleep(0.1)

        for s in series[:n]:
            s.commit_points()



import os
import asyncio
import functools
import random
import time
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import parse_args


class TestBuffer(TestBase):
    title = 'Test buffer object'

    async def _add_points(self):
        for series_name in ['iris', 'db', 'ligo', 'sasha']:
            if series_name not in self.total:
                self.total[series_name] = []
            batches = sum([ord(c) for c in series_name]) % 100
            for i in range(batches):
                npoints = []
                n = int(i**0.5 * 10000 % 5) + 1
                for p in range(n):
                    self.ts += (n + 5000) if i % 2 else (n - 5000)
                    npoints.append([self.ts, i*1000+p])
                self.total[series_name].extend(npoints)
                self.total[series_name].sort()
                await self.client0.insert({series_name: npoints})

    async def _test_equal(self):
        for series_name, points in self.total.items():
            res = await self.client0.query(f'select * from "{series_name}"')
            res = res[series_name]
            self.assertEqual(len(points), len(res))
            self.assertEqual(points, res)

    async def _change_buf_size(self, buffer_size):
        self.client0.close()
        result = await self.server0.stop()
        self.assertTrue(result)
        self.server0.set_buffer_size(self.db, buffer_size)
        await self.server0.start(sleep=20)
        await self.client0.connect()
        res = await self.client0.query('show buffer_size')
        self.assertEqual(res['data'][0]['value'], buffer_size)
        await self._test_equal()

    async def _change_buf_path(self, buffer_path):
        self.client0.close()
        result = await self.server0.stop()
        self.assertTrue(result)
        self.server0.set_buffer_path(self.db, buffer_path)
        await self.server0.start(sleep=20)
        await self.client0.connect()
        res = await self.client0.query('show buffer_path')
        self.assertEqual(res['data'][0]['value'], buffer_path)
        res = await self.client0.query('show open_files')
        self.assertEqual(res['data'][0]['value'], 3)
        res = await self.client0.query(
            f'alter server {self.uuid} set backup_mode true')
        await asyncio.sleep(5)
        res = await self.client0.query('show open_files')
        self.assertEqual(res['data'][0]['value'], 0)
        res = await self.client0.query(
            f'alter server {self.uuid} set backup_mode false')
        await self._test_equal()

    @default_test_setup(1, time_precision='s', compression=False)
    async def run(self):
        await self.client0.connect()

        res = await self.client0.query('show uuid')
        self.uuid = res['data'][0]['value']

        self.ts = 1500000000
        self.total = {}

        await self._add_points()
        await self._test_equal()

        await self._change_buf_path(os.path.join(
            self.server0.dbpath,
            self.db.dbname,
            '../buf/'))

        await self._change_buf_size(8192)

        await self._add_points()
        await self._test_equal()

        await self._change_buf_size(8192)
        await self._change_buf_size(512)

        await self._add_points()
        await self._test_equal()

        await self._change_buf_size(1024)

        await self._change_buf_path(os.path.join(
            self.server0.dbpath,
            self.db.dbname,
            'buf/'))


if __name__ == '__main__':
    parse_args()
    run_test(TestBuffer())



import asyncio
import functools
import random
import time
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import parse_args


class TestServer(TestBase):
    title = 'Test server object'

    Server.SERVER_ADDRESS = 'localhost'
    Server.IP_SUPPORT = 'IPV4ONLY'

    @default_test_setup(4)
    async def run(self):

        await self.client0.connect()

        await self.db.add_pool(self.server1)
        await self.assertIsRunning(self.db, self.client0, timeout=20)
        await asyncio.sleep(5)

        await self.client1.connect()

        for port in (9010, 9011):
            result = await self.client0.query(
                'alter server "localhost:{}" set log_level error'.format(port))
            self.assertEqual(
                result.pop('success_msg'),
                "Successfully set log level to 'error' on 'localhost:{}'."
                .format(port))

        result = await self.client1.query('list servers log_level')
        self.assertEqual(result.pop('servers'), [['error'], ['error']])

        result = await self.client1.query('list servers uuid')

        for uuid in result.pop('servers'):
            result = await self.client0.query(
                'alter server {} set log_level debug'.format(uuid[0]))

        result = await self.client1.query('list servers log_level')
        self.assertEqual(result.pop('servers'), [['debug'], ['debug']])

        result = await self.client0.query('alter servers set log_level info')
        self.assertEqual(
            result.pop('success_msg'),
            "Successfully set log level to 'info' on 2 servers.")

        result = await self.client1.query('list servers log_level')
        self.assertEqual(result.pop('servers'), [['info'], ['info']])

        result = await self.client1.query(
            'list servers active_tasks where active_tasks == 1 and '
            'idle_time >= 0 and idle_percentage <= 100')
        self.assertEqual(result.pop('servers'), [[1], [1]])

        result = await self.client0.query(
            'alter servers where active_handles > 1 set log_level debug')

        result = await self.client1.query('list servers log_level')
        self.assertEqual(result.pop('servers'), [['debug'], ['debug']])

        with self.assertRaisesRegex(
                QueryError,
                "Query error at position 42. Expecting "
                "debug, info, warning, error or critical"):
            await self.client0.query(
                'alter server "localhost:{}" set log_level unknown')

        self.client1.close()
        result = await self.server1.stop()
        self.assertTrue(result)

        self.server1.listen_backend_port = 9111
        self.server1.create()
        await self.server1.start(sleep=20)

        await asyncio.sleep(35)

        result = await self.client0.query('list servers status')
        self.assertEqual(result.pop('servers'), [['running'], ['running']])

        await self.client1.connect()
        result = await self.client1.query('show server')
        self.assertEqual(result.pop('data'), [
            {'name': 'server', 'value': 'localhost:9111'}])

        await self.db.add_replica(self.server2, 1)
        await self.assertIsRunning(self.db, self.client0, timeout=35)

        with self.assertRaisesRegex(
                QueryError,
                "Cannot remove server 'localhost:9010' "
                "because this is the only server for pool 0"):
            await self.client1.query('drop server "localhost:9010"')

        with self.assertRaisesRegex(
                QueryError,
                "Cannot remove server 'localhost:9012' "
                "because the server is still online.*"):
            await self.client1.query('drop server "localhost:9012"')

        result = await self.server1.stop()
        self.assertTrue(result)

        result = await self.server2.stop()
        self.assertTrue(result)

        await self.server1.start(sleep=30)

        result = await self.client1.query('show status')
        self.assertEqual(result.pop('data'), [
            {'name': 'status', 'value': 'running | synchronizing'}])

        result = await self.client0.query('drop server "localhost:9012"')
        self.assertEqual(
            result.pop('success_msg'),
            "Successfully dropped server 'localhost:9012'.")
        self.db.servers.remove(self.server2)

        time.sleep(1)

        for client in (self.client0, self.client1):
            result = await client.query('list servers status')
            self.assertEqual(result.pop('servers'), [['running'], ['running']])

        await self.db.add_replica(self.server3, 1)
        await self.assertIsRunning(self.db, self.client0, timeout=35)

        self.client0.close()
        self.client1.close()


if __name__ == '__main__':
    parse_args()
    run_test(TestServer())



import asyncio
import functools
import random
import time
import math
import re
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import parse_args

LENPOINTS = 70
DATA = {
    'series-001 float': [
        [1471254707, -3.5],
        [1471254705, 1.5],
        [1471254710, -7.3]],
    'series-001 integer': [
        [1471254705, 5],
        [1471254710, -7],
        [1471254708, -3]],
    'series-002 float': [
        [1471254710, -8.3],
        [1471254705, 3.5],
        [1471254707, -2.5]],
    'series-002 integer': [
        [1471254705, 4],
        [1471254708, -1],
        [1471254710, -8]],
    'aggr': [
        [1447250868, 530], [1447251168, 520],
        [1447249033, 531], [1447249337, 534],
        [1447249633, 535], [1447249937, 531],
        [1447250249, 532], [1447250549, 537],
        [1447251449, 54], [1447251749, 54],
        [1447252049, 513], [1447252349, 537],
        [1447253244, 533], [1447253549, 538],
        [1447253849, 534], [1447254149, 532],
        [1447252649, 528], [1447252968, 531],
        [1447254449, 533], [1447254748, 537]],
    'huge': [
        [1471254705, 9223372036854775807],
        [1471254706, 9223372036854775806],
        [1471254707, 9223372036854775805],
        [1471254708, 9223372036854775804]],
    'equal ts': [
        [1471254705, 0], [1471254705, 1], [1471254705, 1],
        [1471254707, 0], [1471254707, 1], [1471254708, 0],
    ],
    'variance': [
        [1471254705, 2.75], [1471254706, 1.75], [1471254707, 1.25],
        [1471254708, 0.25], [1471254709, 0.5], [1471254710, 1.25],
        [1471254711, 3.5]
    ],
    'pvariance': [
        [1471254705, 0.0], [1471254706, 0.25], [1471254707, 0.25],
        [1471254708, 1.25], [1471254709, 1.5], [1471254710, 1.75],
        [1471254711, 2.75], [1471254712, 3.25]
    ],
    'filter': [
        [1471254705, 5],
        [1471254710, -3],
        [1471254715, -7],
        [1471254720, 7]
    ],
    'one': [
        [1471254710, 1]
    ],
    'log': [
        [1471254710, 'log line one'],
        [1471254712, 'log line two'],
        [1471254714, 'another line (three)'],
        [1471254716, 'and yet one more'],
    ],
    'special': [
        [1471254705, 0.1],
        [1471254706, math.nan],
        [1471254707, math.inf],
        [1471254708, -math.inf],
    ]
}


class TestSelect(TestBase):
    title = 'Test select and aggregate functions'

    @default_test_setup(1, compression=False, buffer_size=1024)
    async def run(self):
        await self.client0.connect()

        self.assertEqual(
            await self.client0.insert(DATA),
            {'success_msg': 'Successfully inserted {} point(s).'.format(
                LENPOINTS)})

        for name in DATA:
            DATA[name] = sorted(DATA[name])

        self.assertEqual(
            await self.client0.query(
                'select difference() from "series-001 integer"'),
            {'series-001 integer': [[1471254708, -8], [1471254710, -4]]})

        self.assertEqual(
            await self.client0.query(
                'select difference() => difference() '
                'from "series-001 integer"'),
            {'series-001 integer': [[1471254710, 4]]})

        self.assertEqual(
            await self.client0.query(
                'select difference() => difference() => difference() '
                'from "series-001 integer"'),
            {'series-001 integer': []})

        now = int(time.time())
        self.assertEqual(
            await self.client0.query(
                'select difference({}) from "series-001 integer"'.format(now)),
            {'series-001 integer': [[now, -12]]})

        now = int(time.time())
        self.assertEqual(
            await self.client0.query(
                'select difference({}) from "series-001 integer"'.format(now)),
            {'series-001 integer': [[now, -12]]})

        self.assertEqual(
            await self.client0.query(
                'select * from /series-001.*/ '
                'merge as "median_low" using median_low({})'
                .format(now)),
            {'median_low': [[now, -3.5]]})

        self.assertEqual(
            await self.client0.query(
                'select * from /series-001.*/ '
                'merge as "median_high" using median_high({})'
                .format(now)),
            {'median_high': [[now, -3.0]]})

        self.assertEqual(
            await self.client0.query(
                'select * from /series.*/ '
                'merge as "max" using max(1s)'),
            {'max': [
                [1471254705, 5.0],
                [1471254707, -2.5],
                [1471254708, -1.0],
                [1471254710, -7.0]
            ]})

        # Test all aggregation methods

        self.assertEqual(
            await self.client0.query('select sum(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 2663], [1447254000, 5409], [1447257600, 1602]]})

        self.assertEqual(
            await self.client0.query('select count(1h) from "aggr"'),
            {'aggr': [[1447250400, 5], [1447254000, 12], [1447257600, 3]]})

        self.assertEqual(
            await self.client0.query('select mean(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 532.6],
                [1447254000, 450.75],
                [1447257600, 534.0]]})

        self.assertEqual(
            await self.client0.query('select median(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 532.0],
                [1447254000, 530.5],
                [1447257600, 533.0]]})

        self.assertEqual(
            await self.client0.query('select median_low(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 532], [1447254000, 530], [1447257600, 533]]})

        self.assertEqual(
            await self.client0.query('select median_high(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 532], [1447254000, 531], [1447257600, 533]]})

        self.assertEqual(
            await self.client0.query('select min(1h) from "aggr"'),
            {'aggr': [[1447250400, 531], [1447254000, 54], [1447257600, 532]]})

        self.assertEqual(
            await self.client0.query('select max(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 535], [1447254000, 538], [1447257600, 537]]})

        self.assertAlmostEqual(
            await self.client0.query('select variance(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 3.3],
                [1447254000, 34396.931818181816],
                [1447257600, 7.0]]})

        self.assertAlmostEqual(
            await self.client0.query('select pvariance(1h) from "aggr"'),
            {'aggr': [
                [1447250400, 2.6399999999999997],
                [1447254000, 31530.520833333332],
                [1447257600, 4.666666666666667]]})

        self.assertEqual(
            await self.client0.query('select * from ({}) - ("a", "b")'.format(
                ','.join(['"aggr"'] * 600)
            )),
            {'aggr': DATA['aggr']}
        )

        self.assertEqual(
            await self.client0.query('select difference(1h) from "aggr"'),
            {'aggr': [[1447250400, 1], [1447254000, -3], [1447257600, 5]]})

        self.assertAlmostEqual(
            await self.client0.query('select derivative(1, 1h) from "aggr"'),
            {'aggr': [
                [1447250400, 0.0002777777777777778],
                [1447254000, -0.0008333333333333333],
                [1447257600, 0.001388888888888889]]})

        self.assertEqual(
            await self.client0.query('select filter(>534) from "aggr"'),
            {'aggr': [
                [1447249633, 535],
                [1447250549, 537],
                [1447252349, 537],
                [1447253549, 538],
                [1447254748, 537]]})

        self.assertEqual(
            await self.client0.query(
                'select filter(/l.*/) from * where type == string'),
            {'log': [p for p in DATA['log'] if re.match('l.*', p[1])]})

        self.assertEqual(
            await self.client0.query(
                'select filter(==/l.*/) from * where type == string'),
            {'log': [p for p in DATA['log'] if re.match('l.*', p[1])]})

        self.assertEqual(
            await self.client0.query(
                'select filter(!=/l.*/) from * where type == string'),
            {'log': [p for p in DATA['log'] if not re.match('l.*', p[1])]})

        self.assertEqual(
            await self.client0.query('select limit(300, mean) from "aggr"'),
            {'aggr': DATA['aggr']})

        self.assertEqual(
            await self.client0.query('select limit(1, sum)  from "aggr"'),
            {'aggr': [[1447254748, 9674]]})

        self.assertEqual(
            await self.client0.query('select limit(3, mean) from "aggr"'),
            {'aggr': [
                [1447250938, 532.8571428571429],
                [1447252844, 367.6666666666667],
                [1447254750, 534.0]]})

        self.assertEqual(
            await self.client0.query(
                'select limit(2, max)  from "series-001 float"'),
            {'series-001 float': [[1471254707, 1.5], [1471254713, -7.3]]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select variance(1471254712) from "variance"'),
            {'variance': [[1471254712, 1.3720238095238095]]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select pvariance(1471254715) from "pvariance"'),
            {'pvariance': [[1471254715, 1.25]]})

        self.assertEqual(
            await self.client0.query('select * from "one"'),
            {'one': [[1471254710, 1]]})

        self.assertEqual(
            await self.client0.query('select * from "log"'),
            {'log': DATA['log']})

        self.assertEqual(
            await self.client0.query(
                'select filter(~"log") => filter(!~"one") from "log"'),
            {'log': [DATA['log'][1]]})

        self.assertEqual(
            await self.client0.query(
                'select filter(!=nan) from "special"'),
            {'special': [p for p in DATA['special'] if not math.isnan(p[1])]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select filter(==nan) from "special"'),
            {'special': [p for p in DATA['special'] if math.isnan(p[1])]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select filter(>=nan) from "special"'),
            {'special': [p for p in DATA['special'] if math.isnan(p[1])]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select filter(<=nan) from "special"'),
            {'special': [p for p in DATA['special'] if math.isnan(p[1])]})

        self.assertEqual(
            await self.client0.query(
                'select filter(==inf) from "special"'),
            {'special': [p for p in DATA['special'] if p[1] == math.inf]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select filter(<inf) from "special"'),
            {'special': [p for p in DATA['special'] if p[1] < math.inf]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select filter(>inf) from "special"'),
            {'special': []})

        self.assertEqual(
            await self.client0.query(
                'select filter(==-inf) from "special"'),
            {'special': [p for p in DATA['special'] if p[1] == -math.inf]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select filter(>-inf) from "special"'),
            {'special': [p for p in DATA['special'] if p[1] > -math.inf]})

        self.assertAlmostEqual(
            await self.client0.query(
                'select filter(<-inf) from "special"'),
            {'special': []})

        self.assertEqual(
            await self.client0.query(
                'select filter(~"one") prefix "1-", '
                'filter(~"two") prefix "2-" from "log"'),
            {
                '1-log': [
                    [1471254710, 'log line one'],
                    [1471254716, 'and yet one more']],
                '2-log': [[1471254712, 'log line two']]
            })

        self.assertEqual(
            await self.client0.query('select timeval() from "aggr"'),
            {'aggr': [
                [1447249033, 1447249033],
                [1447249337, 1447249337],
                [1447249633, 1447249633],
                [1447249937, 1447249937],
                [1447250249, 1447250249],
                [1447250549, 1447250549],
                [1447250868, 1447250868],
                [1447251168, 1447251168],
                [1447251449, 1447251449],
                [1447251749, 1447251749],
                [1447252049, 1447252049],
                [1447252349, 1447252349],
                [1447252649, 1447252649],
                [1447252968, 1447252968],
                [1447253244, 1447253244],
                [1447253549, 1447253549],
                [1447253849, 1447253849],
                [1447254149, 1447254149],
                [1447254449, 1447254449],
                [1447254748, 1447254748]]})

        self.assertEqual(
            await self.client0.query('select interval() from "aggr"'),
            {'aggr': [
                [1447249337, 304],
                [1447249633, 296],
                [1447249937, 304],
                [1447250249, 312],
                [1447250549, 300],
                [1447250868, 319],
                [1447251168, 300],
                [1447251449, 281],
                [1447251749, 300],
                [1447252049, 300],
                [1447252349, 300],
                [1447252649, 300],
                [1447252968, 319],
                [1447253244, 276],
                [1447253549, 305],
                [1447253849, 300],
                [1447254149, 300],
                [1447254449, 300],
                [1447254748, 299]]})

        self.assertEqual(
            await self.client0.query('select difference() from "one"'),
            {'one': []})

        with self.assertRaisesRegex(
                QueryError,
                'Regular expressions can only be used with.*'):
            await self.client0.query('select filter(~//) from "log"')

        with self.assertRaisesRegex(
                QueryError,
                'Cannot use a string filter on number type.'):
            await self.client0.query('select filter(//) from "aggr"')

        with self.assertRaisesRegex(
                QueryError,
                r'Cannot use mean\(\) on string type\.'):
            await self.client0.query('select mean(1w) from "log"')

        with self.assertRaisesRegex(
                QueryError,
                r'Group by time must be an integer value larger than zero\.'):
            await self.client0.query('select mean(0) from "aggr"')

        with self.assertRaisesRegex(
                QueryError,
                r'Limit must be an integer value larger than zero\.'):
            await self.client0.query('select limit(6 - 6, mean) from "aggr"')

        with self.assertRaisesRegex(
                QueryError,
                r'Cannot use a string filter on number type\.'):
            await self.client0.query(
                'select * from "aggr" '
                'merge as "t" using filter("0")')

        with self.assertRaisesRegex(
                QueryError,
                r'Cannot use difference\(\) on string type\.'):
            await self.client0.query('select difference() from "log"')

        with self.assertRaisesRegex(
                QueryError,
                r'Cannot use derivative\(\) on string type\.'):
            await self.client0.query('select derivative(6, 3) from "log"')

        with self.assertRaisesRegex(
                QueryError,
                r'Cannot use derivative\(\) on string type\.'):
            await self.client0.query('select derivative() from "log"')

        with self.assertRaisesRegex(
                QueryError,
                r'Overflow detected while using sum\(\)\.'):
            await self.client0.query('select sum(now) from "huge"')

        with self.assertRaisesRegex(
                QueryError,
                'Max depth reached in \'where\' expression!'):
            await self.client0.query(
                'select * from "aggr" where ((((((length > 1))))))')

        with self.assertRaisesRegex(
                QueryError,
                'Cannot compile regular expression.*'):
            await self.client0.query(
                'select * from /(bla/')

        with self.assertRaisesRegex(
                QueryError,
                'Memory allocation error or maximum recursion depth reached.'):
            await self.client0.query(
                'select * from {}"aggr"{}'.format(
                    '(' * 501,
                    ')' * 501))

        with self.assertRaisesRegex(
                    QueryError,
                    'Query too long.'):
            await self.client0.query('select * from "{}"'.format('a' * 65535))

        with self.assertRaisesRegex(
                    QueryError,
                    'Error while merging points. Make sure the destination '
                    'series name is valid.'):
            await self.client0.query(
                'select * from "aggr", "huge" merge as ""')

        self.assertEqual(
            await self.client0.query(
                'select min(2h) prefix "min-", max(1h) prefix "max-" '
                'from /.*/ where type == integer and name != "filter" '
                'and name != "one" and name != "series-002 integer" '
                'merge as "int_min_max" using median_low(1) => difference()'),
            {
                'max-int_min_max': [
                    [1447254000, 3], [1447257600, -1], [1471255200, -532]],
                'min-int_min_max': [
                    [1447257600, -477], [1471255200, -54]]})

        await self.client0.query('select derivative() from "equal ts"')

        self.assertEqual(
            await self.client0.query('select first() from *'),
            {k: [v[0]] for k, v in DATA.items()})

        self.assertEqual(
            await self.client0.query('select last() from *'),
            {k: [v[-1]] for k, v in DATA.items()})

        self.assertEqual(
            await self.client0.query('select count() from *'),
            {k: [[v[-1][0], len(v)]] for k, v in DATA.items()})

        self.assertEqual(
            await self.client0.query('select mean() from "aggr"'),
            {'aggr': [[
                DATA['aggr'][-1][0],
                sum([x[1] for x in DATA['aggr']]) / len(DATA['aggr'])]]})

        self.assertAlmostEqual(
            await self.client0.query('select stddev() from "aggr"'),
            {'aggr': [[
                DATA['aggr'][-1][0],
                147.07108914792838]]})

        self.assertAlmostEqual(
            await self.client0.query('select stddev(1h) from "aggr"'),
            {"aggr": [
                [1447250400, 1.8165902124584952],
                [1447254000, 185.46409846162092],
                [1447257600, 2.6457513110645907]]})

        # test prefix, suffex
        result = await self.client0.query(
                'select sum(1d) prefix "sum-" suffix "-sum", '
                'min(1d) prefix "minimum-", '
                'max(1d) suffix "-maximum" from "aggr"')

        self.assertIn('sum-aggr-sum', result)
        self.assertIn('minimum-aggr', result)
        self.assertIn('aggr-maximum', result)

        await self.client0.query('alter database set select_points_limit 10')
        with self.assertRaisesRegex(
                QueryError,
                'Query has reached the maximum number of selected points.*'):
            await self.client0.query(
                'select * from /.*/')
        await self.client0.query(
            'alter database set select_points_limit 1000000')

        self.client0.close()


if __name__ == '__main__':
    parse_args()
    run_test(TestSelect())



import asyncio
import functools
import random
import time
from testing import Client
from testing import default_test_setup
from testing import gen_data
from testing import gen_points
from testing import gen_series
from testing import InsertError
from testing import PoolError
from testing import QueryError
from testing import run_test
from testing import Series
from testing import Server
from testing import ServerError
from testing import SiriDB
from testing import TestBase
from testing import UserAuthError
from testing import parse_args


DATA = {
    'log': [
        [1471254710, 'log line one'],
        [1471254712, 'log line two'],
        [1471254714, 'log line three'],
        [1471254715, 'log line four'],
        [1471254716, 'log line five'],
        [1471254718, 'another line (six)'],
        [1471254720, 'and yet one more, this is log line seven']
    ],
    'utf16': [
        [1471254710, ''],
        [1471254712, ''],
        [1471254714, '  '],
        [1471254715, ' '],
        [1471254716, ' '],
        [1471254718, 'ԉ']
    ],
    'long_log': [
        [1471254710, '''
time series is a series of data points indexed (or listed or graphed) in time order. Most commonly, a time series is a sequence taken at successive equally spaced points in time. Thus it is a sequence of discrete-time data. Examples of time series are heights of ocean tides, counts of sunspots, and the daily closing value of the Dow Jones Industrial Average.
Time series are very frequently plotted via line charts. Time series are used in statistics, signal processing, pattern recognition, econometrics, mathematical finance, weather forecasting, earthquake prediction, electroencephalography, control engineering, astronomy, communications engineering, and largely in any domain of applied science and engineering which involves temporal measurements.
Time series analysis comprises methods for analyzing time series data in order to extract meaningful statistics and other characteristics of the data. Time series forecasting is the use of a model to predict future values based on previously observed values. While regression analysis is often employed in such a way as to test theories that the current values of one or more independent time series affect the current value of another time series, this type of analysis of time series is not called "time series analysis", which focuses on comparing values of a single time series or multiple dependent time series at different points in time.[1] Interrupted time series analysis is the analysis of interventions on a single time series.
Time series data have a natural temporal ordering. This makes time series analysis distinct from cross-sectional studies, in which there is no natural ordering of the observations (e.g. explaining people's wages by reference to their respective education levels, where the individuals' data could be entered in any order). Time series analysis is also distinct from spatial data analysis where the observations typically relate to geographical locations (e.g. accounting for house prices by the location as well as the intrinsic characteristics of the houses). A stochastic model for a time series will generally reflect the fact that observations close together in time will be more closely related than observations further apart. In addition, time series models will often make use of the natural one-way ordering of time so that values for a given period will be expressed as deriving in some way from past values, rather than from future values (see time reversibility.)
Time series analysis can be applied to real-valued, continuous data, discrete numeric data, or discrete symbolic data (i.e. sequences of characters, such as letters and words in the English language[2]).']
'''],  # nopep8
        [1471254712, '''
Methods for time series analysis may be divided into two classes: frequency-domain methods and time-domain methods. The former include spectral analysis and wavelet analysis; the latter include auto-correlation and cross-correlation analysis. In the time domain, correlation and analysis can be made in a filter-like manner using scaled correlation, thereby mitigating the need to operate in the frequency domain.
Additionally, time series analysis techniques may be divided into parametric and non-parametric methods. The parametric approaches assume that the underlying stationary stochastic process has a certain structure which can be described using a small number of parameters (for example, using an autoregressive or moving average model). In these approaches, the task is to estimate the parameters of the model that describes the stochastic process. By contrast, non-parametric approaches explicitly estimate the covariance or the spectrum of the process without assuming that the process has any particular structure.
Methods of time series analysis may also be divided into linear and non-linear, and univariate and multivariate.
'''],  # nopep8
        [1471254714, '''
Curve fitting
Main article: Curve fitting
Curve fitting[5][6] is the process of constructing a curve, or mathematical function, that has the best fit to a series of data points,[7] possibly subject to constraints.[8][9] Curve fitting can involve either interpolation,[10][11] where an exact fit to the data is required, or smoothing,[12][13] in which a "smooth" function is constructed that approximately fits the data. A related topic is regression analysis,[14][15] which focuses more on questions of statistical inference such as how much uncertainty is present in a curve that is fit to data observed with random errors. Fitted curves can be used as an aid for data visualization,[16][17] to infer values of a function where no data are available,[18] and to summarize the relationships among two or more variables.[19] Extrapolation refers to the use of a fitted curve beyond the range of the observed data,[20] and is subject to a degree of uncertainty[21] since it may reflect the method used to construct the curve as much as it reflects the observed data.
The construction of economic time series involves the estimation of some components for some dates by interpolation between values ("benchmarks") for earlier and later dates. Interpolation is estimation of an unknown quantity between two known quantities (historical data), or drawing conclusions about missing information from the available information ("reading between the lines").[22] Interpolation is useful where the data surrounding the missing data is available and its trend, seasonality, and longer-term cycles are known. This is often done by using a related series known for all relevant dates.[23] Alternatively polynomial interpolation or spline interpolation is used where piecewise polynomial functions are fit into time intervals such that they fit smoothly together. A different problem which is closely related to interpolation is the approximation of a complicated function by a simple function (also called regression).The main difference between regression and interpolation is that polynomial regression gives a single polynomial that models the entire data set. Spline interpolation, however, yield a piecewise continuous function composed of many polynomials to model the data set.
Extrapolation is the process of estimating, beyond the original observation range, the value of a variable on the basis of its relationship with another variable. It is similar to interpolation, which produces estimates between known observations, but extrapolation is subject to greater uncertainty and a higher risk of producing meaningless results.

'''],  # nopep8
        [1471254715, '''
Function approximation
Main article: Function approximation
In general, a function approximation problem asks us to select a function among a well-defined class that closely matches ("approximates") a target function in a task-specific way. One can distinguish two major classes of function approximation problems: First, for known target functions approximation theory is the branch of numerical analysis that investigates how certain known functions (for example, special functions) can be approximated by a specific class of functions (for example, polynomials or rational functions) that often have desirable properties (inexpensive computation, continuity, integral and limit values, etc.).
Second, the target function, call it g, may be unknown; instead of an explicit formula, only a set of points (a time series) of the form (x, g(x)) is provided. Depending on the structure of the domain and codomain of g, several techniques for approximating g may be applicable. For example, if g is an operation on the real numbers, techniques of interpolation, extrapolation, regression analysis, and curve fitting can be used. If the codomain (range or target set) of g is a finite set, one is dealing with a classification problem instead. A related problem of online time series approximation[24] is to summarize the data in one-pass and construct an approximate representation that can support a variety of time series queries with bounds on worst-case error.
To some extent the different problems (regression, classification, fitness approximation) have received a unified treatment in statistical learning theory, where they are viewed as supervised learning problems.
'''],  # nopep8
        [1471254716, '''
Prediction and forecasting
In statistics, prediction is a part of statistical inference. One particular approach to such inference is known as predictive inference, but the prediction can be undertaken within any of the several approaches to statistical inference. Indeed, one description of statistics is that it provides a means of transferring knowledge about a sample of a population to the whole population, and to other related populations, which is not necessarily the same as prediction over time. When information is transferred across time, often to specific points in time, the process is known as forecasting.
Fully formed statistical models for stochastic simulation purposes, so as to generate alternative versions of the time series, representing what might happen over non-specific time-periods in the future
Simple or fully formed statistical models to describe the likely outcome of the time series in the immediate future, given knowledge of the most recent outcomes (forecasting).
Forecasting on time series is usually done using automated statistical software packages and programming languages, such as R, S, SAS, SPSS, Minitab, pandas (Python) and many others.
Forecasting on large scale data is done using Spark which has spark-ts as a third party package.
Classification
Main article: Statistical classification
Assigning time series pattern to a specific category, for example identify a word based on series of hand movements in sign language.
Signal estimation
See also: Signal processing and Estimation theory
This approach is based on harmonic analysis and filtering of signals in the frequency domain using the Fourier transform, and spectral density estimation, the development of which was significantly accelerated during World War II by mathematician Norbert Wiener, electrical engineers Rudolf E. Kálmán, Dennis Gabor and others for filtering signals from noise and predicting signal values at a certain point in time. See Kalman filter, Estimation theory, and Digital signal processing
'''],  # nopep8
        [1471254718, '''
Segmentation
Main article: Time-series segmentation
Splitting a time-series into a sequence of segments. It is often the case that a time-series can be represented as a sequence of individual segments, each with its own characteristic properties. For example, the audio signal from a conference call can be partitioned into pieces corresponding to the times during which each person was speaking. In time-series segmentation, the goal is to identify the segment boundary points in the time-series, and to characterize the dynamical properties associated with each segment. One can approach this problem using change-point detection, or by modeling the time-series as a more sophisticated system, such as a Markov jump linear system.
Models
Models for time series data can have many forms and represent different stochastic processes. When modeling variations in the level of a process, three broad classes of practical importance are the autoregressive (AR) models, the integrated (I) models, and the moving average (MA) models. These three classes depend linearly on previous data points.[25] Combinations of these ideas produce autoregressive moving average (ARMA) and autoregressive integrated moving average (ARIMA) models. The autoregressive fractionally integrated moving average (ARFIMA) model generalizes the former three. Extensions of these classes to deal with vector-valued data are available under the heading of multivariate time-series models and sometimes the preceding acronyms are extended by including an initial "V" for "vector", as in VAR for vector autoregression. An additional set of extensions of these models is available for use where the observed time-series is driven by some "forcing" time-series (which may not have a causal effect on the observed series): the distinction from the multivariate case is that the forcing series may be deterministic or under the experimenter's control. For these models, the acronyms are extended with a final "X" for "exogenous".
Non-linear dependence of the level of a series on previous data points is of interest, partly because of the possibility of producing a chaotic time series. However, more importantly, empirical investigations can indicate the advantage of using predictions derived from non-linear models, over those from linear models, as for example in nonlinear autoregressive exogenous models. Further references on nonlinear time series analysis: (Kantz and Schreiber),[26] and (Abarbanel)[27]
Among other types of non-linear time series models, there are models to represent the changes of variance over time (heteroskedasticity). These models represent autoregressive conditional heteroskedasticity (ARCH) and the collection comprises a wide variety of representation (GARCH, TARCH, EGARCH, FIGARCH, CGARCH, etc.). Here changes in variability are related to, or predicted by, recent past values of the observed series. This is in contrast to other possible representations of locally varying variability, where the variability might be modelled as being driven by a separate time-varying process, as in a doubly stochastic model.
In recent work on model-free analyses, wavelet transform based methods (for example locally stationary wavelets and wavelet decomposed neural networks) have gained favor. Multiscale (often referred to as multiresolution) techniques decompose a given time series, attempting to illustrate time dependence at multiple scales. See also Markov switching multifractal (MSMF) techniques for modeling volatility evolution.
A Hidden Markov model (HMM) is a statistical Markov model in which the system being modeled is assumed to be a Markov process with unobserved (hidden) states. An HMM can be considered as the simplest dynamic Bayesian network. HMM models are widely used in speech recognition, for translating a time series of spoken words into text.
'''],  # nopep8
        [1471254720, '''
Time
The flow of sand in an hourglass can be used to measure the passage of time. It also concretely represents the present as being between the past and the future.
Time is the indefinite continued progress of existence and events that occur in apparently irreversible succession from the past through the present to the future.[1][2][3] Time is a component quantity of various measurements used to sequence events, to compare the duration of events or the intervals between them, and to quantify rates of change of quantities in material reality or in the conscious experience.[4][5][6][7] Time is often referred to as a fourth dimension, along with three spatial dimensions.[8]
Time has long been an important subject of study in religion, philosophy, and science, but defining it in a manner applicable to all fields without circularity has consistently eluded scholars.[2][6][7][9][10][11] Nevertheless, diverse fields such as business, industry, sports, the sciences, and the performing arts all incorporate some notion of time into their respective measuring systems.[12][13][14]
Two contrasting viewpoints on time divide prominent philosophers. One view is that time is part of the fundamental structure of the universe – a dimension independent of events, in which events occur in sequence. Isaac Newton subscribed to this realist view, and hence it is sometimes referred to as Newtonian time.[15][16] The opposing view is that time does not refer to any kind of "container" that events and objects "move through", nor to any entity that "flows", but that it is instead part of a fundamental intellectual structure (together with space and number) within which humans sequence and compare events. This second view, in the tradition of Gottfried Leibniz[17] and Immanuel Kant,[18][19] holds that time is neither an event nor a thing, and thus is not itself measurable nor can it be travelled.
Time in physics is unambiguously operationally defined as "what a clock reads".[6][17][20] See Units of Time. Time is one of the seven fundamental physical quantities in both the International System of Units and International System of Quantities. Time is used to define other quantities – such as velocity – so defining time in terms of such quantities would result in circularity of definition.[21] An operational definition of time, wherein one says that observing a certain number of repetitions of one or another standard cyclical event (such as the passage of a free-swinging pendulum) constitutes one standard unit such as the second, is highly useful in the conduct of both advanced experiments and everyday affairs of life. The operational definition leaves aside the question whether there is something called time, apart from the counting activity just mentioned, that flows and that can be measured. Investigations of a single continuum called spacetime bring questions about space into questions about time, questions that have their roots in the works of early students of natural philosophy.
Temporal measurement has occupied scientists and technologists, and was a prime motivation in navigation and astronomy. Periodic events and periodic motion have long served as standards for units of time. Examples include the apparent motion of the sun across the sky, the phases of the moon, the swing of a pendulum, and the beat of a heart. Currently, the international unit of time, the second, is defined by measuring the electronic transition frequency of caesium atoms (see below). Time is also of significant social importance, having economic value ("time is money") as well as personal value, due to an awareness of the limited time in each day and in human life spans.]
Temporal measurement and history
Generally speaking, methods of temporal measurement, or chronometry, take two distinct forms: the calendar, a mathematical tool for organising intervals of time,[22] and the clock, a physical mechanism that counts the passage of time. In day-to-day life, the clock is consulted for periods less than a day whereas the calendar is consulted for periods longer than a day. Increasingly, personal electronic devices display both calendars and clocks simultaneously. The number (as on a clock dial or calendar) that marks the occurrence of a specified event as to hour or date is obtained by counting from a fiducial epoch – a central reference point.
History of the calendar
Main article: Calendar
Artifacts from the Paleolithic suggest that the moon was used to reckon time as early as 6,000 years ago.[23] Lunar calendars were among the first to appear, either 12 or 13 lunar months (either 354 or 384 days). Without intercalation to add days or months to some years, seasons quickly drift in a calendar based solely on twelve lunar months. Lunisolar calendars have a thirteenth month added to some years to make up for the difference between a full year (now known to be about 365.24 days) and a year of just twelve lunar months. The numbers twelve and thirteen came to feature prominently in many cultures, at least partly due to this relationship of months to years. Other early forms of calendars originated in Mesoamerica, particularly in ancient Mayan civilization. These calendars were religiously and astronomically based, with 18 months in a year and 20 days in a month, plus five epagomenal days at the end of the year.[24]
The reforms of Julius Caesar in 45 BC put the Roman world on a solar calendar. This Julian calendar was faulty in that its intercalation still allowed the astronomical solstices and equinoxes to advance against it by about 11 minutes per year. Pope Gregory XIII introduced a correction in 1582; the Gregorian calendar was only slowly adopted by different nations over a period of centuries, but it is now the most commonly used calendar around the world, by far.
During the French Revolution, a new clock and calendar were invented in attempt to de-Christianize time and create a more rational system in order to replace the Gregorian calendar. The French Republican Calendar's days consisted of ten hours of a hundred minutes of a hundred seconds, which marked a deviation from the 12-based duodecimal system used in many other devices by many cultures. The system was later abolished in 1806.[25]
History of time measurement devices
Horizontal sundial in Taganrog
An old kitchen clock
Main article: History of timekeeping devices
See also: Clock
A large variety of devices have been invented to measure time. The study of these devices is called horology.
An Egyptian device that dates to c.1500 BC, similar in shape to a bent T-square, measured the passage of time from the shadow cast by its crossbar on a nonlinear rule. The T was oriented eastward in the mornings. At noon, the device was turned around so that it could cast its shadow in the evening direction.[26]
A sundial uses a gnomon to cast a shadow on a set of markings calibrated to the hour. The position of the shadow marks the hour in local time. The idea to separate the day into smaller parts is credited to Egyptians because of their sundials, which operated on a duodecimal system. The importance of the number 12 is due the number of lunar cycles in a year and the number of stars used to count the passage of night.[27]
The most precise timekeeping device of the ancient world was the water clock, or clepsydra, one of which was found in the tomb of Egyptian pharaoh Amenhotep I (1525–1504 BC). They could be used to measure the hours even at night, but required manual upkeep to replenish the flow of water. The Ancient Greeks and the people from Chaldea (southeastern Mesopotamia) regularly maintained timekeeping records as an essential part of their astronomical observations. Arab inventors and engineers in particular made improvements on the use of water clocks up to the Middle Ages.[28] In the 11th century, Chinese inventors and engineers invented the first mechanical clocks driven by an escapement mechanism.
The hourglass uses the flow of sand to measure the flow of time. They were used in navigation. Ferdinand Magellan used 18 glasses on each ship for his circumnavigation of the globe (1522).[29] Incense sticks and candles were, and are, commonly used to measure time in temples and churches across the globe. Waterclocks, and later, mechanical clocks, were used to mark the events of the abbeys and monasteries of the Middle Ages. Richard of Wallingford (1292–1336), abbot of St. Alban's abbey, famously built a mechanical clock as an astronomical orrery about 1330.[30][31] Great advances in accurate time-keeping were made by Galileo Galilei and especially Christiaan Huygens with the invention of pendulum driven clocks along with the invention of the minute hand by Jost Burgi.[32]
The English word clock probably comes from the Middle Dutch word klocke which, in turn, derives from the medieval Latin word clocca, which ultimately derives from Celtic and is cognate with French, Latin, and German words that mean bell. The passage of the hours at sea were marked by bells, and denoted the time (see ship's bell). The hours were marked by bells in abbeys as well as at sea.
Clocks can range from watches, to more exotic varieties such as the Clock of the Long Now. They can be driven by a variety of means, including gravity, springs, and various forms of electrical power, and regulated by a variety of means such as a pendulum.
The English word clock probably comes from the Middle Dutch word klocke which, in turn, derives from the medieval Latin word clocca, which ultimately derives from Celtic and is cognate with French, Latin, and German words that mean bell. The passage of the hours at sea were marked by bells, and denoted the time (see ship's bell). The hours were marked by bells in abbeys as well as at sea.
Chip-scale atomic clocks, such as this one unveiled in 2004, are expected to greatly improve GPS location.[33]
Clocks can range from watches, to more exotic varieties such as the Clock of the Long Now. They can be driven by a variety of means, including gravity, springs, and various forms of electrical power, and regulated by a variety of means such as a pendulum.
Alarm clocks first appeared in ancient Greece around 250 BC with a water clock that would set off a whistle. This idea was later mechanized by Levi Hutchins and Seth E. Thomas
A chronometer is a portable timekeeper that meets certain precision standards. Initially, the term was used to refer to the marine chronometer, a timepiece used to determine longitude by means of celestial navigation, a precision firstly achieved by John Harrison. More recently, the term has also been applied to the chronometer watch, a watch that meets precision standards set by the Swiss agency COSC.
The most accurate timekeeping devices are atomic clocks, which are accurate to seconds in many millions of years,[34] and are used to calibrate other clocks and timekeeping instruments. Atomic clocks use the frequency of electronic transitions in certain atoms to measure the second. One of the most common atoms used is caesium, most modern atomic clocks probe caesium with microwaves to determine the frequency of these electron vibrations.[35] Since 1967, the International System of Measurements bases its unit of time, the second, on the properties of caesium atoms. SI defines the second as 9,192,631,770 cycles of the radiation that corresponds to the transition between two electron spin energy levels of the ground state of the 133Cs atom.
Today, the Global Positioning System in coordination with the Network Time Protocol can be used to synchronize timekeeping systems across the globe.
Definitions and standards
The Mean Solar Time system defines the second as 1/86,400 of the mean solar day, which is the year-average of the solar day. The solar day is the time interval between two successive solar noons, i.e., the time interval between two successive passages of the Sun across the local meridian. The local meridian is an imaginary line that runs from celestial north pole to celestial south pole passing directly over the head of the observer. At the local meridian the Sun reaches its highest point on its daily arc across the sky.
In 1874 the British Association for the Advancement of Science introduced the CGS (centimetre/gramme/second system) combining fundamental units of length, mass and time. The second is "elastic", because tidal friction is slowing the earth's rotation rate. For use in calculating ephemerides of celestial motion, therefore, in 1952 astronomers introduced the "ephemeris second", currently defined as the fraction 1/31,556,925.9747 of the tropical year for 1900 January 0 at 12 hours ephemeris time.
The CGS system has been superseded by the Système international. The SI base unit for time is the SI second. The International System of Quantities, which incorporates the SI, also defines larger units of time equal to fixed integer multiples of one second (1 s), such as the minute, hour and day. These are not part of the SI, but may be used alongside the SI. Other units of time such as the month and the year are not equal to fixed multiples of 1 s, and instead exhibit significant variations in duration.
The official SI definition of the second is as follows: The second is the duration of 9,192,631,770 periods of the radiation corresponding to the transition between the two hyperfine levels of the ground state of the caesium 133 atom.
At its 1997 meeting, the CIPM affirmed that this definition refers to a caesium atom in its ground state at a temperature of 0 K.
The current definition of the second, coupled with the current definition of the metre, is based on the special theory of relativity, which affirms our spacetime to be a Minkowski space. The definition of the second in mean solar time, however, is unchanged.
World time
While in theory, the concept of a single worldwide universal time-scale may have been conceived of many centuries ago, in practicality the technical ability to create and maintain such a time-scale did not become possible until the mid-19th century. The timescale adopted was Greenwich Mean Time, created in 1847. A few countries have replaced it with Coordinated Universal Time, UTC.
History of the development of UTC
With the advent of the industrial revolution, a greater understanding and agreement on the nature of time itself became increasingly necessary and helpful. In 1847 in Britain, Greenwich Mean Time (GMT) was first created for use by the British railways, the British navy, and the British shipping industry. Using telescopes, GMT was calibrated to the mean solar time at the Royal Observatory, Greenwich in the UK.
As international commerce continued to increase throughout Europe, in order to achieve a more efficiently functioning modern society, an agreed upon, and highly accurate international standard of time measurement became necessary. In order to find or determine such a time-standard, three steps had to be followed:
An internationally agreed upon time-standard had to be defined.
This new time-standard then had to be consistently and accurately measured.
The new time-standard then had to be freely shared and distributed around the world.
The development of what is now known as UTC time came about historically as an effort which first began as a collaboration between 41 nations, officially agreed to and signed at the International Meridian Conference, in Washington D.C. in 1884. At this conference, the local mean solar time at the Royal Observatory, Greenwich in England was chosen to define the "universal day", counted from 0 hours at Greenwich mean midnight. This agreed with the civil Greenwich Mean Time used on the island of Great Britain since 1847. In contrast astronomical GMT began at mean noon, i.e. astronomical day X began at noon of civil day X. The purpose of this was to keep one night's observations under one date. The civil system was adopted as of 0 hours (civil) 1 January 1925. Nautical GMT began 24 hours before astronomical GMT, at least until 1805 in the Royal Navy, but persisted much later elsewhere because it was mentioned at the 1884 conference. In 1884, the Greenwich meridian was used for two-thirds of all charts and maps as their Prime Meridian.[42]
Among the 41 nations represented at the conference, the advanced time-technologies that had already come into use in Britain were fundamental components of the agreed upon method of arriving at a universal and agreed upon international time. In 1928 Greenwich Mean Time was rebranded for scientific purposes by the International Astronomical Union as Universal Time (UT). This was to avoid confusion with the previous system where the day had begun at noon. As the general public had always begun the day at midnight the timescale continued to be presented to them as Greenwich Mean Time. By 1956, universal time had been split into various versions – UT2, which smoothed for polar motion and seasonal effects, was presented to the public as Greenwich Mean Time. Later, UT1 (which smooths only for polar motion) became the default form of UT used by astronomers and hence the form used in navigation, sunrise and sunset and moonrise and moonset tables where the name Greenwich Mean Time continues to be employed. Greenwich Mean Time is also the preferred method of describing the timescale used by legislators. Even to the present day, UT is still based on an international telescopic system. Observations at the Greenwich Observatory itself ceased in 1954, though the location is still used as the basis for the coordinate system. Because the rotational period of Earth is not perfectly constant, the duration of a second would vary if calibrated to a telescope-based standard like GMT, where the second is defined as 1/86 400 of the mean solar day.
For the better part of the first century following the "International Meridian Conference," until 1960, the methods and definitions of time-keeping that had been laid out at the conference proved to be adequate to meet time tracking needs of science. Still, with the advent of the "electronic revolution" in the latter half of the 20th century, the technologies that had been available at the time of the Convention of the Metre proved to be in need of further refinement in order to meet the needs of the ever-increasing precision that the "electronic revolution" had begun to require.
The ephemeris second
An invariable second (the "ephemeris second") had been defined, use of which removed the errors in ephemerides resulting from the use of the variable mean solar second as the time argument. In 1960 this ephemeris second was made the basis of the "coordinated universal time" which was being derived from atomic clocks. It is a specified fraction of the mean tropical year as at 1900 and, being based on historical telescope observations, corresponds roughly to the mean solar second of the early nineteenth century.[43]
The SI second
In 1967 a further step was taken with the introduction of the SI second, essentially the ephemeris second as measured by atomic clocks and formally defined in atomic terms.[44] The SI second (Standard Internationale second) is based directly on the measurement of the atomic-clock observation of the frequency oscillation of caesium atoms. It is the basis of all atomic timescales, e.g. coordinated universal time, GPS time, International Atomic Time, etc. Atomic clocks do not measure nuclear decay rates, which is a common misconception, but rather measure a certain natural vibrational frequency of caesium-133.[45] Coordinated universal time is subject to one constraint which does not affect the other atomic timescales. As it has been adopted as the civil timescale by some countries (most countries have opted to retain mean solar time) it is not permitted to deviate from GMT by more than 0.9 second. This is achieved by the occasional insertion of a leap second.
Current application of UTC
Most countries use mean solar time. Australia, Canada (Quebec only), Colombia, France, Germany, New Zealand, Papua New Guinea (Bougainville only), Paraguay, Portugal, Switzerland, the United States and Venezuela use UTC. However, UTC is widely used by the scientific community in countries where mean solar time is official. UTC time is based on the SI second, which was first defined in 1967, and is based on the use of atomic clocks. Some other less used but closely related time-standards include International Atomic Time (TAI), Terrestrial Time, and Barycentric Dynamical Time.
Between 1967 and 1971, UTC was periodically adjusted by fractional amounts of a second in order to adjust and refine for variations in mean solar time, with which it is aligned. After 1 January 1972, UTC time has been defined as being offset from atomic time by a whole number of seconds, changing only when a leap second is added to keep radio-controlled clocks synchronized with the rotation of the Earth.
The Global Positioning System also broadcasts a very precise time signal worldwide, along with instructions for converting GPS time to UTC. GPS-time is based on, and regularly synchronized with or from, UTC-time.
Earth is split up into a number of time zones. Most time zones are exactly one hour apart, and by convention compute their local time as an offset from GMT. For example, time zones at sea are based on GMT. In many locations (but not at sea) these offsets vary twice yearly due to daylight saving time transitions.
Philosophy
Main articles: Philosophy of space and time and Temporal finitism
Time's mortal aspect is personified in this bronze statue by Charles van der Stappen.
Two distinct viewpoints on time divide many prominent philosophers. One view is that time is part of the fundamental structure of the universe, a dimension in which events occur in sequence. Sir Isaac Newton subscribed to this realist view, and hence it is sometimes referred to as Newtonian time.[16] An opposing view is that time does not refer to any kind of actually existing dimension that events and objects "move through", nor to any entity that "flows", but that it is instead an intellectual concept (together with space and number) that enables humans to sequence and compare events.[56] This second view, in the tradition of Gottfried Leibniz[17] and Immanuel Kant,[18][19] holds that space and time "do not exist in and of themselves, but ... are the product of the way we represent things", because we can know objects only as they appear to us.
Furthermore, it may be that there is a subjective component to time, but whether or not time itself is "felt", as a sensation, or is a judgment, is a matter of debate.[2][6][7][57][58]
The Vedas, the earliest texts on Indian philosophy and Hindu philosophy dating back to the late 2nd millennium BC, describe ancient Hindu cosmology, in which the universe goes through repeated cycles of creation, destruction and rebirth, with each cycle lasting 4,320 million years.[59] Ancient Greek philosophers, including Parmenides and Heraclitus, wrote essays on the nature of time.[60] Plato, in the Timaeus, identified time with the period of motion of the heavenly bodies. Aristotle, in Book IV of his Physica defined time as 'number of movement in respect of the before and after'.[61]
In Book 11 of his Confessions, St. Augustine of Hippo ruminates on the nature of time, asking, "What then is time? If no one asks me, I know: if I wish to explain it to one that asketh, I know not." He begins to define time by what it is not rather than what it is,[62] an approach similar to that taken in other negative definitions. However, Augustine ends up calling time a "distention" of the mind (Confessions 11.26) by which we simultaneously grasp the past in memory, the present by attention, and the future by expectation.
This view is shared by Abrahamic faiths as they believe time started by creation, therefore the only thing being infinite is God and everything else, including time, is finite.
Isaac Newton believed in absolute space and absolute time; Leibniz believed that time and space are relational.[63] The differences between Leibniz's and Newton's interpretations came to a head in the famous Leibniz–Clarke correspondence.
Time is not an empirical concept. For neither co-existence nor succession would be perceived by us, if the representation of time did not exist as a foundation a priori. Without this presupposition we could not represent to ourselves that things exist together at one and the same time, or at different times, that is, contemporaneously, or in succession.
Immanuel Kant, Critique of Pure Reason (1781), trans. Vasilis Politis (London: Dent., 1991), p.54.
Immanuel Kant, in the Critique of Pure Reason, described time as an a priori intuition that allows us (together with the other a priori intuition, space) to comprehend sense experience.[64] With Kant, neither space nor time are conceived as substances, but rather both are elements of a systematic mental framework that necessarily structures the experiences of any rational agent, or observing subject. Kant thought of time as a fundamental part of an abstract conceptual framework, together with space and number, within which we sequence events, quantify their duration, and compare the motions of objects. In this view, time does not refer to any kind of entity that "flows," that objects "move through," or that is a "container" for events. Spatial measurements are used to quantify the extent of and distances between objects, and temporal measurements are used to quantify the durations of and between events. Time was designated by Kant as the purest possible schema of a pure concept or category.
Henri Bergson believed that time was neither a real homogeneous medium nor a mental construct, but possesses what he referred to as Duration. Duration, in Bergson's view, was creativity and memory as an essential component of reality.[65]
According to Martin Heidegger we do not exist inside time, we are time. Hence, the relationship to the past is a present awareness of having been, which allows the past to exist in the present. The relationship to the future is the state of anticipating a potential possibility, task, or engagement. It is related to the human propensity for caring and being concerned, which causes "being ahead of oneself" when thinking of a pending occurrence. Therefore, this concern for a potential occurrence also allows the future to exist in the present. The present becomes an experience, which is qualitative instead of quantitative. Heidegger seems to think this is the way that a linear relationship with time, or temporal existence, is broken or transcended.[66] We are not stuck in sequential time. We are able to remember the past and project into the future – we have a kind of random access to our representation of temporal existence; we can, in our thoughts, step out of (ecstasis) sequential time.
Time as "unreal"
In 5th century BC Greece, Antiphon the Sophist, in a fragment preserved from his chief work On Truth, held that: "Time is not a reality (hypostasis), but a concept (noêma) or a measure (metron)." Parmenides went further, maintaining that time, motion, and change were illusions, leading to the paradoxes of his follower Zeno.[68] Time as an illusion is also a common theme in Buddhist thought.[69][70]
J. M. E. McTaggart's 1908 The Unreality of Time argues that, since every event has the characteristic of being both present and not present (i.e., future or past), that time is a self-contradictory idea (see also The flow of time).
These arguments often center on what it means for something to be unreal. Modern physicists generally believe that time is as real as space – though others, such as Julian Barbour in his book The End of Time, argue that quantum equations of the universe take their true form when expressed in the timeless realm containing every possible now or momentary configuration of the universe, called 'platonia' by Barbour.[71]
A modern philosophical theory called presentism views the past and the future as human-mind interpretations of movement instead of real parts of time (or "dimensions") which coexist with the present. This theory rejects the existence of all direct interaction with the past or the future, holding only the present as tangible. This is one of the philosophical arguments against time travel. This contrasts with eternalism (all time: present, past and future, is real) and the growing block theory (the present and the past are real, but the future is not).
Physical Definition
Until Einstein's reinterpretation of the physical concepts associated with time and space, time was considered to be the same everywhere in the universe, with all observers measuring the same time interval for any event.[72] Non-relativistic classical mechanics is based on this Newtonian idea of time.
Einstein, in his special theory of relativity,[73] postulated the constancy and finiteness of the speed of light for all observers. He showed that this postulate, together with a reasonable definition for what it means for two events to be simultaneous, requires that distances appear compressed and time intervals appear lengthened for events associated with objects in motion relative to an inertial observer.
The theory of special relativity finds a convenient formulation in Minkowski spacetime, a mathematical structure that combines three dimensions of space with a single dimension of time. In this formalism, distances in space can be measured by how long light takes to travel that distance, e.g., a light-year is a measure of distance, and a meter is now defined in terms of how far light travels in a certain amount of time. Two events in Minkowski spacetime are separated by an invariant interval, which can be either space-like, light-like, or time-like. Events that have a time-like separation cannot be simultaneous in any frame of reference, there must be a temporal component (and possibly a spatial one) to their separation. Events that have a space-like separation will be simultaneous in some frame of reference, and there is no frame of reference in which they do not have a spatial separation. Different observers may calculate different distances and different time intervals between two events, but the invariant interval between the events is independent of the observer (and his or her velocity).
Classical mechanics
In non-relativistic classical mechanics, Newton's concept of "relative, apparent, and common time" can be used in the formulation of a prescription for the synchronization of clocks. Events seen by two different observers in motion relative to each other produce a mathematical concept of time that works sufficiently well for describing the everyday phenomena of most people's experience. In the late nineteenth century, physicists encountered problems with the classical understanding of time, in connection with the behavior of electricity and magnetism. Einstein resolved these problems by invoking a method of synchronizing clocks using the constant, finite speed of light as the maximum signal velocity. This led directly to the result that observers in motion relative to one another measure different elapsed times for the same event.
Spacetime
Main article: Spacetime
Time has historically been closely related with space, the two together merging into spacetime in Einstein's special relativity and general relativity. According to these theories, the concept of time depends on the spatial reference frame of the observer, and the human perception as well as the measurement by instruments such as clocks are different for observers in relative motion. For example, if a spaceship carrying a clock flies through space at (very nearly) the speed of light, its crew does not notice a change in the speed of time on board their vessel because everything traveling at the same speed slows down at the same rate (including the clock, the crew's thought processes, and the functions of their bodies). However, to a stationary observer watching the spaceship fly by, the spaceship appears flattened in the direction it is traveling and the clock on board the spaceship appears to move very slowly.
On the other hand, the crew on board the spaceship also perceives the observer as slowed down and flattened along the spaceship's direction of travel, because both are moving at very nearly the speed of light relative to each other. Because the outside universe appears flattened to the spaceship, the crew perceives themselves as quickly traveling between regions of space that (to the stationary observer) are many light years apart. This is reconciled by the fact that the crew's perception of time is different from the stationary observer's; what seems like seconds to the crew might be hundreds of years to the stationary observer. In either case, however, causality remains unchanged: the past is the set of events that can send light signals to an entity and the future is the set of events to which an entity can send light signals
Time dilation
Relativity of simultaneity: Event B is simultaneous with A in the green reference frame, but it occurred before in the blue frame, and occurs later in the red frame.
Main article: Time dilation
Einstein showed in his thought experiments that people travelling at different speeds, while agreeing on cause and effect, measure different time separations between events, and can even observe different chronological orderings between non-causally related events. Though these effects are typically minute in the human experience, the effect becomes much more pronounced for objects moving at speeds approaching the speed of light. Subatomic particles exist for a well known average fraction of a second in a lab relatively at rest, but when travelling close to the speed of light they are measured to travel farther and exist for much longer than when at rest. According to the special theory of relativity, in the high-speed particle's frame of reference, it exists, on the average, for a standard amount of time known as its mean lifetime, and the distance it travels in that time is zero, because its velocity is zero. Relative to a frame of reference at rest, time seems to "slow down" for the particle. Relative to the high-speed particle, distances seem to shorten. Einstein showed how both temporal and spatial dimensions can be altered (or "warped") by high-speed motion.
Einstein (The Meaning of Relativity): "Two events taking place at the points A and B of a system K are simultaneous if they appear at the same instant when observed from the middle point, M, of the interval AB. Time is then defined as the ensemble of the indications of similar clocks, at rest relative to K, which register the same simultaneously."
Einstein wrote in his book, Relativity, that simultaneity is also relative, i.e., two events that appear simultaneous to an observer in a particular inertial reference frame need not be judged as simultaneous by a second observer in a different inertial frame of reference.
Relativistic time versus Newtonian time
Views of spacetime along the world line of a rapidly accelerating observer in a relativistic universe. The events ("dots") that pass the two diagonal lines in the bottom half of the image (the past light cone of the observer in the origin) are the events visible to the observer.
The animations visualise the different treatments of time in the Newtonian and the relativistic descriptions. At the heart of these differences are the Galilean and Lorentz transformations applicable in the Newtonian and relativistic theories, respectively.
In the figures, the vertical direction indicates time. The horizontal direction indicates distance (only one spatial dimension is taken into account), and the thick dashed curve is the spacetime trajectory ("world line") of the observer. The small dots indicate specific (past and future) events in spacetime.
The slope of the world line (deviation from being vertical) gives the relative velocity to the observer. Note how in both pictures the view of spacetime changes when the observer accelerates.
In the Newtonian description these changes are such that time is absolute:[77] the movements of the observer do not influence whether an event occurs in the 'now' (i.e., whether an event passes the horizontal line through the observer).
However, in the relativistic description the observability of events is absolute: the movements of the observer do not influence whether an event passes the "light cone" of the observer. Notice that with the change from a Newtonian to a relativistic description, the concept of absolute time is no longer applicable: events move up-and-down in the figure depending on the acceleration of the observer.
Time appears to have a direction – the past lies behind, fixed and immutable, while the future lies ahead and is not necessarily fixed. Yet for the most part the laws of physics do not specify an arrow of time, and allow any process to proceed both forward and in reverse. This is generally a consequence of time being modelled by a parameter in the system being analysed, where there is no "proper time": the direction of the arrow of time is sometimes arbitrary. Examples of this include the cosmological arrow of time, which points away from the Big Bang, CPT symmetry, and the radiative arrow of time, caused by light only travelling forwards in time (see light cone). In particle physics, the violation of CP symmetry implies that there should be a small counterbalancing time asymmetry to preserve CPT symmetry as stated above. The standard description of measurement in quantum mechanics is also time asymmetric (see Measurement in quantum mechanics). The second law of thermodynamics states that entropy must increase over time (see Entropy). This can be in either direction – Brian Greene theorizes that, according to the equations, the change in entropy occurs symmetrically whether going forward or backward in time. So entropy tends to increase in either direction, and our current low-entropy universe is a statistical aberration, in the similar manner as tossing a coin often enough that eventually heads will result ten times in a row. However, this theory is not supported empirically in local experiment.[78]
''']  # nopep8
    ]}


class TestLog(TestBase):
    title = 'Test log'

    @default_test_setup(1, compression=True, optimize_interval=10)
    async def run(self):
        await self.client0.connect()

        n = 0
        for p in DATA.values():
            n += len(p)

        self.assertEqual(
            await self.client0.insert(DATA),
            {'success_msg': 'Successfully inserted {} point(s).'.format(n)})

        await self.test_data()

        # wait for optimization
        await asyncio.sleep(30)

        await self.test_data()

        self.client0.close()

        # stop and restart
        result = await self.server0.stop()
        self.assertTrue(result)

        await self.server0.start(sleep=20)

        await self.client0.connect()

        await self.test_data()

    async def test_data(self):
        self.assertEqual(
            await self.client0.query('select * from "log"'),
            {'log': DATA['log']})

        self.assertEqual(
            await self.client0.query('select * from "utf16"'),
            {'utf16': DATA['utf16']})

        self.assertEqual(
            await self.client0.query('select * from "long_log"'),
            {'long_log': DATA['long_log']})


if __name__ == '__main__':
    parse_args()
    run_test(TestLog())
